@@ -1,5 +1,5 @@
 /*
-  $Id: sample4.c,v 1.1 2003/08/17 07:12:45 rocky Exp $
+  $Id: sample4.c,v 1.3 2003/10/03 01:42:47 rocky Exp $
 
   Copyright (C) 2003 Rocky Bernstein <rocky@panix.com>
   
@@ -22,6 +22,9 @@
    A slightly improved sample3 program: we handle cdio logging and
    take an optional CD-location.
 */
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
@@ -47,7 +50,7 @@ print_analysis(cdio_analysis_t cdio_analysis,
 	       track_t num_tracks, track_t first_track_num, CdIo *cdio)
 {
   switch(CDIO_FSTYPE(fs)) {
-  case CDIO_FS_NO_DATA:
+  case CDIO_FS_AUDIO:
     break;
   case CDIO_FS_ISO_9660:
     printf("CD-ROM with ISO 9660 filesystem");
@@ -111,7 +114,7 @@ print_analysis(cdio_analysis_t cdio_analysis,
     printf("CD-Plus/Extra   ");
   if (fs & CDIO_FS_ANAL_BOOTABLE)
     printf("bootable CD   ");
-  if (fs & CDIO_FS_ANAL_VIDEOCDI && num_audio == 0) {
+  if (fs & CDIO_FS_ANAL_VIDEOCD && num_audio == 0) {
     printf("Video CD   ");
   }
   if (fs & CDIO_FS_ANAL_SVCD)
@@ -148,6 +151,7 @@ main(int argc, const char *argv[])
 
   if (NULL == cdio) {
     printf("Problem in trying to find a driver.\n\n");
+    return 1;
   }
 
   first_track_num = cdio_get_first_track_num(cdio);
