@@ -1,5 +1,5 @@
 /* -*- c -*-
-    $Id: cdio.h,v 1.17 2003/08/31 14:26:06 rocky Exp $
+    $Id: cdio.h,v 1.31 2003/10/04 23:11:50 rocky Exp $
 
     Copyright (C) 2001 Herbert Valerio Riedel <hvr@gnu.org>
     Copyright (C) 2003 Rocky Bernstein <rocky@panix.com>
@@ -74,6 +74,7 @@ extern "C" {
     DRIVER_FREEBSD, 
     DRIVER_LINUX,
     DRIVER_SOLARIS, 
+    DRIVER_OSX, 
     DRIVER_WIN32,
     DRIVER_BINCUE, /* Prefer bincue over nrg when both exist */
     DRIVER_NRG,    
@@ -114,6 +115,12 @@ extern "C" {
    */
   void cdio_destroy (CdIo *obj);
 
+  /*!
+    Free device list returned by cdio_get_devices or
+    cdio_get_devices_with_cap.
+  */
+  void cdio_free_device_list (char * device_list[]);
+
   /*!
     Return the value associatied with key. NULL is returned if obj is NULL
     or "key" does not exist.
@@ -121,14 +128,49 @@ extern "C" {
   const char * cdio_get_arg (const CdIo *obj,  const char key[]);
 
   /*!
-    Return a string containing the default CD device if none is specified.
-    if CdIo is NULL (we haven't initialized a specific device driver), 
+    Return an array of device names in search_devices that have at
+    least the capabilities listed by cap.  If search_devices is NULL,
+    then we'll search all possible CD drives.  
+    
+    If "any" is set false then every capability listed in the extended
+    portion of capabilities (i.e. not the basic filesystem) must be
+    satisified. If "any" is set true, then if any of the capabilities
+    matches, we call that a success.
+
+    To find a CD-drive of any type, use the mask CDIO_FS_MATCH_ALL.
+  
+    NULL is returned if we couldn't get a default device.
+  */
+  char ** cdio_get_devices_with_cap (char* search_devices[],
+				     cdio_fs_anal_t capabilities, bool any);
+
+  /*!Return an array of device names. If you want a specific
+    devices, dor a driver give that device, if you want hardware
+    devices, give DRIVER_DEVICE and if you want all possible devices,
+    image drivers and hardware drivers give DRIVER_UNKNOWN.
+    
+    NULL is returned if we couldn't return a list of devices.
+  */
+  char ** cdio_get_devices (driver_id_t driver);
+
+  /*!
+    Return a string containing the default CD device.
+    if obj is NULL (we haven't initialized a specific device driver), 
     then find a suitable one and return the default device for that.
     
     NULL is returned if we couldn't get a default device.
   */
   char * cdio_get_default_device (const CdIo *obj);
 
+  /*!
+    Return the media catalog number (MCN) from the CD or NULL if there
+    is none or we don't have the ability to get it.
+
+    Note: string is malloc'd so caller has to free() the returned
+    string when done with it.
+  */
+  char *cdio_get_mcn (const CdIo *obj);
+
   /*!
     Return a string containing the name of the driver in use.
     if CdIo is NULL (we haven't initialized a specific device driver), 
@@ -203,57 +245,67 @@ extern "C" {
     Returns (off_t) -1 on error. 
     Similar to (if not the same as) libc's lseek()
   */
-  off_t cdio_lseek(CdIo *obj, off_t offset, int whence);
+  off_t cdio_lseek(const CdIo *obj, off_t offset, int whence);
     
   /*!
     Reads into buf the next size bytes.
     Returns -1 on error. 
     Similar to (if not the same as) libc's read()
   */
-  ssize_t cdio_read(CdIo *obj, void *buf, size_t size);
+  ssize_t cdio_read(const CdIo *obj, void *buf, size_t size);
     
   /*!
    Reads a audio sector from cd device into data starting
    from lsn. Returns 0 if no error. 
  */
-  int cdio_read_audio_sector (CdIo *obj, void *buf, lsn_t lsn);
+  int cdio_read_audio_sector (const CdIo *obj, void *buf, lsn_t lsn);
+
+  /*!
+   Reads a audio sector from cd device into data starting
+   from lsn. Returns 0 if no error. 
+ */
+  int cdio_read_audio_sectors (const CdIo *obj, void *buf, lsn_t lsn,
+			       unsigned int nblocks);
 
   /*!
    Reads a single mode1 sector from cd device into data starting
    from lsn. Returns 0 if no error. 
  */
-  int cdio_read_mode1_sector (CdIo *obj, void *buf, lsn_t lsn, bool is_form2);
+  int cdio_read_mode1_sector (const CdIo *obj, void *buf, lsn_t lsn, 
+			      bool is_form2);
 
   /*!
    Reads nblocks of mode1 sectors from cd device into data starting
    from lsn. Returns 0 if no error. 
  */
-  int cdio_read_mode1_sectors (CdIo *obj, void *buf, lsn_t lsn, bool is_form2,
-			       unsigned int num_sectors);
+  int cdio_read_mode1_sectors (const CdIo *obj, void *buf, lsn_t lsn, 
+			       bool is_form2, unsigned int num_sectors);
 
   /*!
    Reads a single mode2 sector from cd device into data starting
    from lsn. Returns 0 if no error. 
  */
-  int cdio_read_mode2_sector (CdIo *obj, void *buf, lsn_t lsn, bool is_form2);
+  int cdio_read_mode2_sector (const CdIo *obj, void *buf, lsn_t lsn, 
+			      bool is_form2);
 
   /*!
     Reads nblocks of mode2 sectors from cd device into data starting
     from lsn.
     Returns 0 if no error. 
   */
-  int cdio_read_mode2_sectors (CdIo *obj, void *buf, lsn_t lsn, bool is_form2, 
-			       unsigned int num_sectors);
+  int cdio_read_mode2_sectors (const CdIo *obj, void *buf, lsn_t lsn, 
+			       bool is_form2, unsigned int num_sectors);
   
   /*!
     Set the arg "key" with "value" in the source device.
+    0 is returned if no error was found, and nonzero if there as an error.
   */
   int cdio_set_arg (CdIo *obj, const char key[], const char value[]);
   
   /*!
     Return the size of the CD in logical block address (LBA) units.
   */
-  uint32_t cdio_stat_size (CdIo *obj);
+  uint32_t cdio_stat_size (const CdIo *obj);
 
   /*!
     Initialize CD Reading and control routines. Should be called first.
@@ -266,6 +318,7 @@ extern "C" {
   bool cdio_have_freebsd (void);
   bool cdio_have_linux   (void);
   bool cdio_have_solaris (void);
+  bool cdio_have_osx     (void);
   bool cdio_have_win32   (void);
   bool cdio_have_nrg     (void);
   bool cdio_have_bincue  (void);
@@ -293,6 +346,8 @@ extern "C" {
   
   char * cdio_get_default_device_bincue(void);
 
+  char **cdio_get_devices_bincue(void);
+
   /*! CD routines. Source is the some sort of device.
 
      NULL is returned on error.
@@ -312,6 +367,8 @@ extern "C" {
   
   char * cdio_get_default_device_bsdi(void);
 
+  char **cdio_get_devices_bsdi(void);
+  
   /*! BSDI CD-reading routines. 
      NULL is returned on error.
   */
@@ -319,12 +376,16 @@ extern "C" {
   
   char * cdio_get_default_device_freebsd(void);
 
+  char **cdio_get_devices_freebsd(void);
+  
   /*! Linux CD-reading routines. 
      NULL is returned on error.
    */
   CdIo * cdio_open_linux (const char *source_name);
 
   char * cdio_get_default_device_linux(void);
+
+  char **cdio_get_devices_linux(void);
   
   /*! Solaris CD-reading routines. 
      NULL is returned on error.
@@ -333,6 +394,17 @@ extern "C" {
   
   char * cdio_get_default_device_solaris(void);
   
+  char **cdio_get_devices_solaris(void);
+  
+  /*! Darwin OS X CD-reading routines. 
+     NULL is returned on error.
+   */
+  CdIo * cdio_open_osx (const char *source_name);
+
+  char * cdio_get_default_device_osx(void);
+  
+  char **cdio_get_devices_osx(void);
+  
   /*! Win32 CD-reading routines. 
      NULL is returned on error.
    */
@@ -340,6 +412,8 @@ extern "C" {
   
   char * cdio_get_default_device_win32(void);
 
+  char **cdio_get_devices_win32(void);
+  
   /*! Nero CD disk-image routines. 
      NULL is returned on error.
    */
@@ -347,6 +421,8 @@ extern "C" {
   
   char * cdio_get_default_device_nrg(void);
 
+  char **cdio_get_devices_nrg(void);
+
   /*! Return corresponding BIN file if cue_name is a cue file or NULL
     if not a CUE file.
   */
