@@ -1,5 +1,5 @@
 /*
-    $Id: types.h,v 1.5 2003/08/31 01:01:40 rocky Exp $
+    $Id: types.h,v 1.7 2003/09/28 17:14:20 rocky Exp $
 
     Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>
     Copyright (C) 2002,2003 Rocky Bernstein <rocky@panix.com>
@@ -200,7 +200,7 @@ extern "C" {
   /* The type of an Logical Sector Number. */
   typedef uint32_t lsn_t;
   
-  /* The type of an track number. */
+  /* The type of an track number 0..99. */
   typedef uint8_t track_t;
   
   /*! 
@@ -218,6 +218,8 @@ extern "C" {
   */
 #define CDIO_INVALID_LSN   0xFFFFFFFF
 
+typedef int cdio_fs_anal_t;
+
 #ifdef __cplusplus
 }
 #endif /* __cplusplus */
