@@ -1,5 +1,5 @@
 /*
-    $Id: _cdio_bincue.c,v 1.24 2003/09/01 22:51:50 rocky Exp $
+    $Id: _cdio_bincue.c,v 1.36 2003/10/03 07:55:00 rocky Exp $
 
     Copyright (C) 2001 Herbert Valerio Riedel <hvr@gnu.org>
     Copyright (C) 2002,2003 Rocky Bernstein <rocky@panix.com>
@@ -24,23 +24,41 @@
    (*.cue).
 */
 
-#ifdef HAVE_CONFIG_H
-# include "config.h"
-#endif
+static const char _rcsid[] = "$Id: _cdio_bincue.c,v 1.36 2003/10/03 07:55:00 rocky Exp $";
 
-static const char _rcsid[] = "$Id: _cdio_bincue.c,v 1.24 2003/09/01 22:51:50 rocky Exp $";
-
-#include <stdio.h>
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
+#include "cdio_assert.h"
+#include "cdio_private.h"
+#include "_cdio_stdio.h"
 
 #include <cdio/logging.h>
 #include <cdio/sector.h>
 #include <cdio/util.h>
-#include "cdio_assert.h"
-#include "cdio_private.h"
-#include "_cdio_stdio.h"
+
+#ifdef HAVE_STDIO_H
+#include <stdio.h>
+#endif
+#ifdef HAVE_STDLIB_H
+#include <stdlib.h>
+#endif
+#ifdef HAVE_STRING_H
+#include <string.h>
+#endif
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
+#ifdef HAVE_GLOB_H
+#include <glob.h>
+#endif
+#include <ctype.h>
+
+/* FIXME: should put in a common definition somewhere. */
+#ifdef HAVE_BZERO
+#define BZERO(ptr, size) bzero(ptr, size)
+#elif  HAVE_MEMSET
+#define BZERO(ptr, size) memset(ptr, 0, size)
+#else 
+  Error -- you need either memset or bzero
+#endif
 
 /* reader */
 
@@ -76,6 +94,7 @@ typedef struct {
   bool sector_2336;              /* Playstation (PSX) uses 2336-byte sectors */
 
   char         *cue_name;
+  char         *mcn;             /* Media catalog number. */
   track_info_t  tocent[100];     /* entry info for each track */
   track_t       total_tracks;    /* number of tracks in image */
   track_t       first_track_num; /* track number of first track */
@@ -83,7 +102,7 @@ typedef struct {
 } _img_private_t;
 
 static bool     _cdio_image_read_cue (_img_private_t *_obj);
-static uint32_t _cdio_stat_size (void *user_data);
+static uint32_t _cdio_stat_size (void *env);
 
 /*!
   Initialize image structures.
@@ -156,9 +175,9 @@ _cdio_init (_img_private_t *_obj)
   information in each sector.
 */
 static off_t
-_cdio_lseek (void *user_data, off_t offset, int whence)
+_cdio_lseek (void *env, off_t offset, int whence)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   /* real_offset is the real byte offset inside the disk image
      The number below was determined empirically. I'm guessing
@@ -203,9 +222,9 @@ _cdio_lseek (void *user_data, off_t offset, int whence)
    boundaries.
 */
 static ssize_t
-_cdio_read (void *user_data, void *data, size_t size)
+_cdio_read (void *env, void *data, size_t size)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
   char *p = data;
   ssize_t final_size=0;
@@ -254,9 +273,9 @@ _cdio_read (void *user_data, void *data, size_t size)
    Return the size of the CD in logical block address (LBA) units.
  */
 static uint32_t 
-_cdio_stat_size (void *user_data)
+_cdio_stat_size (void *env)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   long size;
   int blocksize = _obj->sector_2336 
     ? M2RAW_SECTOR_SIZE : CDIO_CD_FRAMESIZE_RAW;
@@ -302,18 +321,22 @@ _cdio_image_read_cue (_img_private_t *_obj)
 
   _obj->total_tracks=0;
   _obj->first_track_num=1;
+  _obj->mcn=NULL;
+  
   while ((fgets(line, MAXLINE, fp)) != NULL) {
-    char *s=NULL;
+    char s[80];
     char *p;
     /*printf("Retrieved line of length %zu :\n", read);
       printf("%s", line); */
     for (p=line; isspace(*p); p++) ;
-    if (1==sscanf(p, "FILE \"%a[^\"]", &s)) {
+    if (1==sscanf(p, "FILE \"%80s[^\"]", s)) {
       /* Should expand file name based on cue file basename.
       free(_obj->bin_file);
-      _obj->bin_file = s;
+      _obj->bin_file = strdup(s);
       */
       /* printf("Found file name %s\n", s); */
+    } else if (1==sscanf(p, "CATALOG %80s", s)) {
+      _obj->mcn = strdup(s);
     } else if (2==sscanf(p, "TRACK %d MODE2/%d", &track_num, &blocksize)) {
       track_info_t  *this_track=&(_obj->tocent[_obj->total_tracks]);
       this_track->track_num   = track_num;
@@ -414,17 +437,20 @@ _cdio_image_read_cue (_img_private_t *_obj)
 	  /* Figure out number of sectors for previous track */
 	  track_info_t  *prev_track=&(_obj->tocent[_obj->total_tracks-2]);
 	  if ( this_track->start_lba < prev_track->start_lba ) {
-	    cdio_warn ("track %d at LBA %d starts before track %d at LBA %d", 
-		       _obj->total_tracks,   this_track->start_lba, 
-		       _obj->total_tracks-1, prev_track->start_lba);
+	    cdio_warn("track %d at LBA %lu starts before track %d at LBA %lu", 
+		     _obj->total_tracks,   
+		      (unsigned long int) this_track->start_lba, 
+		      _obj->total_tracks-1, 
+		      (unsigned long int) prev_track->start_lba);
 	    prev_track->sec_count = 0;
 	  } else if ( this_track->start_lba >= prev_track->start_lba 
 		      + CDIO_PREGAP_SECTORS ) {
 	    prev_track->sec_count = this_track->start_lba - 
 	      prev_track->start_lba - CDIO_PREGAP_SECTORS ;
 	  } else {
-	    cdio_warn ("%d fewer than pregap (%d) sectors in track %d", 
-		       this_track->start_lba - prev_track->start_lba,
+	    cdio_warn ("%lu fewer than pregap (%d) sectors in track %d", 
+		       (long unsigned int) 
+		         this_track->start_lba - prev_track->start_lba,
 		       CDIO_PREGAP_SECTORS,
 		       _obj->total_tracks-1);
 	    /* Include pregap portion in sec_count. Maybe the pregap
@@ -448,22 +474,35 @@ _cdio_image_read_cue (_img_private_t *_obj)
    from lsn. Returns 0 if no error. 
  */
 static int
-_cdio_read_audio_sector (void *user_data, void *data, lsn_t lsn)
+_cdio_read_audio_sectors (void *env, void *data, lsn_t lsn, 
+			  unsigned int nblocks)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   int ret;
 
   _cdio_init (_obj);
 
-  ret = cdio_stream_seek (_obj->gen.data_source, lsn * CDIO_CD_FRAMESIZE_RAW, 
-			  SEEK_SET);
-  if (ret!=0) return ret;
+  /* Why the adjustment of 272, I don't know. It seems to work though */
+  if (lsn != 0) {
+    ret = cdio_stream_seek (_obj->gen.data_source, 
+			    (lsn * CDIO_CD_FRAMESIZE_RAW) - 272, SEEK_SET);
+    if (ret!=0) return ret;
 
-  ret = cdio_stream_read (_obj->gen.data_source, data, CDIO_CD_FRAMESIZE_RAW, 
-			  1);
-  if (ret==0) return ret;
+    ret = cdio_stream_read (_obj->gen.data_source, data, 
+			    CDIO_CD_FRAMESIZE_RAW, nblocks);
+  } else {
+    /* We need to pad out the first 272 bytes with 0's */
+    BZERO(data, 272);
+    
+    ret = cdio_stream_seek (_obj->gen.data_source, 0, SEEK_SET);
 
-  return 0;
+    if (ret!=0) return ret;
+
+    ret = cdio_stream_read (_obj->gen.data_source, (uint8_t *) data+272, 
+			    CDIO_CD_FRAMESIZE_RAW - 272, nblocks);
+  }
+
+  return ret;
 }
 
 /*!
@@ -471,10 +510,10 @@ _cdio_read_audio_sector (void *user_data, void *data, lsn_t lsn)
    from lsn. Returns 0 if no error. 
  */
 static int
-_cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn, 
-		    bool mode2_form2)
+_cdio_read_mode2_sector (void *env, void *data, lsn_t lsn, 
+			 bool mode2_form2)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   int ret;
   char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
   int blocksize = _obj->sector_2336 
@@ -507,10 +546,10 @@ _cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn,
    Returns 0 if no error. 
  */
 static int
-_cdio_read_mode2_sectors (void *user_data, void *data, uint32_t lsn, 
+_cdio_read_mode2_sectors (void *env, void *data, uint32_t lsn, 
 		     bool mode2_form2, unsigned int nblocks)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   int i;
   int retval;
 
@@ -535,13 +574,29 @@ _cdio_read_mode2_sectors (void *user_data, void *data, uint32_t lsn,
 #define free_if_notnull(obj) \
   if (NULL != obj) free(obj);
 
+static void 
+_cdio_bincue_free (void *env) {
+  _img_private_t *_obj = env;
+
+  if (NULL == _obj) return;
+  free_if_notnull(_obj->mcn);
+  cdio_generic_stream_free(_obj);
+}
+
+/*!
+  Set the arg "key" with "value" in the source device.
+  Currently "source" to set the source device in I/O operations 
+  is the only valid key.
+
+  0 is returned if no error was found, and nonzero if there as an error.
+*/
 /*!
   Set the device to use in I/O operations.
 */
 static int
-_cdio_set_arg (void *user_data, const char key[], const char value[])
+_cdio_set_arg (void *env, const char key[], const char value[])
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   if (!strcmp (key, "source"))
     {
@@ -580,9 +635,9 @@ _cdio_set_arg (void *user_data, const char key[], const char value[])
   Return the value associated with the key "arg".
 */
 static const char *
-_cdio_get_arg (void *user_data, const char key[])
+_cdio_get_arg (void *env, const char key[])
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   if (!strcmp (key, "source")) {
     return _obj->gen.source_name;
@@ -593,12 +648,39 @@ _cdio_get_arg (void *user_data, const char key[])
 }
 
 /*!
-  Return a string containing the default VCD device if none is specified.
+  Return an array of strings giving possible NRG disk images.
+ */
+char **
+cdio_get_devices_bincue (void)
+{
+  char **drives = NULL;
+  unsigned int num_files=0;
+#ifdef HAVE_GLOB_H
+  unsigned int i;
+  glob_t globbuf;
+  globbuf.gl_offs = 0;
+  glob("*.cue", GLOB_DOOFFS, NULL, &globbuf);
+  for (i=0; i<globbuf.gl_pathc; i++) {
+    cdio_add_device_list(&drives, globbuf.gl_pathv[i], &num_files);
+  }
+  globfree(&globbuf);
+#else
+  cdio_add_device_list(&drives, DEFAULT_CDIO_DEVICE, &num_files);
+#endif /*HAVE_GLOB_H*/
+  cdio_add_device_list(&drives, NULL, &num_files);
+  return drives;
+}
+
+/*!
+  Return a string containing the default CD device.
  */
 char *
 cdio_get_default_device_bincue(void)
 {
-  return strdup(DEFAULT_CDIO_DEVICE);
+  char **drives = cdio_get_devices_nrg();
+  char *drive = (drives[0] == NULL) ? NULL : strdup(drives[0]);
+  cdio_free_device_list(drives);
+  return drive;
 }
 
 /*!
@@ -606,15 +688,33 @@ cdio_get_default_device_bincue(void)
   CDIO_INVALID_TRACK is returned on error.
 */
 static track_t
-_cdio_get_first_track_num(void *user_data) 
+_cdio_get_first_track_num(void *env) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   _cdio_init (_obj);
 
   return _obj->first_track_num;
 }
 
+/*!
+  Return the media catalog number (MCN) from the CD or NULL if there
+  is none or we don't have the ability to get it.
+
+  Note: string is malloc'd so caller has to free() the returned
+  string when done with it.
+  */
+static char *
+_cdio_get_mcn(void *env)
+{
+  _img_private_t *_obj = env;
+  
+  _cdio_init (_obj);
+
+  if (NULL == _obj->mcn) return NULL;
+  return strdup(_obj->mcn);
+}
+
 /*! 
   Return the number of tracks in the current medium.
   If no cuesheet is available, We fake it an just say there's
@@ -622,9 +722,9 @@ _cdio_get_first_track_num(void *user_data)
   CDIO_INVALID_TRACK is returned on error.
 */
 static track_t
-_cdio_get_num_tracks(void *user_data) 
+_cdio_get_num_tracks(void *env) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   _cdio_init (_obj);
 
   return _obj->have_cue && _obj->total_tracks > 0 ? _obj->total_tracks : 1;
@@ -635,9 +735,9 @@ _cdio_get_num_tracks(void *user_data)
   CDIO_INVALID_TRACK is returned on error.
 */
 static track_format_t
-_cdio_get_track_format(void *user_data, track_t track_num) 
+_cdio_get_track_format(void *env, track_t track_num) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   if (!_obj->gen.init) _cdio_init(_obj);
 
@@ -656,9 +756,9 @@ _cdio_get_track_format(void *user_data, track_t track_num)
   FIXME: there's gotta be a better design for this and get_track_format?
 */
 static bool
-_cdio_get_track_green(void *user_data, track_t track_num) 
+_cdio_get_track_green(void *env, track_t track_num) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   if (!_obj->gen.init) _cdio_init(_obj);
 
@@ -669,16 +769,16 @@ _cdio_get_track_green(void *user_data, track_t track_num)
 }
 
 /*!  
-  Return the starting MSF (minutes/secs/frames) for track number
-  track_num in obj.  Tracks numbers start at 1.
+  Return the starting LSN track number
+  track_num in obj.  Track numbers start at 1.
   The "leadout" track is specified either by
   using track_num LEADOUT_TRACK or the total tracks+1.
-  1 is returned on error.
+  False is returned if there is no track entry.
 */
 static lba_t
-_cdio_get_track_lba(void *user_data, track_t track_num)
+_cdio_get_track_lba(void *env, track_t track_num)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   _cdio_init (_obj);
 
   if (track_num == CDIO_CDROM_LEADOUT_TRACK) track_num = _obj->total_tracks+1;
@@ -697,9 +797,9 @@ _cdio_get_track_lba(void *user_data, track_t track_num)
 
 */
 static bool
-_cdio_get_track_msf(void *user_data, track_t track_num, msf_t *msf)
+_cdio_get_track_msf(void *env, track_t track_num, msf_t *msf)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   _cdio_init (_obj);
 
   if (NULL == msf) return false;
@@ -784,10 +884,11 @@ cdio_open_common (_img_private_t **_data)
 {
   cdio_funcs _funcs = {
     .eject_media        = cdio_generic_bogus_eject_media,
-    .free               = cdio_generic_stream_free,
+    .free               = _cdio_bincue_free,
     .get_arg            = _cdio_get_arg,
     .get_default_device = cdio_get_default_device_bincue,
     .get_first_track_num= _cdio_get_first_track_num,
+    .get_mcn            = _cdio_get_mcn,
     .get_num_tracks     = _cdio_get_num_tracks,
     .get_track_format   = _cdio_get_track_format,
     .get_track_green    = _cdio_get_track_green,
@@ -795,7 +896,7 @@ cdio_open_common (_img_private_t **_data)
     .get_track_msf      = _cdio_get_track_msf,
     .lseek              = _cdio_lseek,
     .read               = _cdio_read,
-    .read_audio_sector  = _cdio_read_audio_sector,
+    .read_audio_sectors = _cdio_read_audio_sectors,
     .read_mode2_sector  = _cdio_read_mode2_sector,
     .read_mode2_sectors = _cdio_read_mode2_sectors,
     .set_arg            = _cdio_set_arg,
