@@ -1,5 +1,5 @@
 /*
-    $Id: _cdio_linux.c,v 1.16 2003/09/01 15:11:36 rocky Exp $
+    $Id: _cdio_linux.c,v 1.27 2003/10/03 01:26:52 rocky Exp $
 
     Copyright (C) 2001 Herbert Valerio Riedel <hvr@gnu.org>
     Copyright (C) 2002,2003 Rocky Bernstein <rocky@panix.com>
@@ -27,7 +27,7 @@
 # include "config.h"
 #endif
 
-static const char _rcsid[] = "$Id: _cdio_linux.c,v 1.16 2003/09/01 15:11:36 rocky Exp $";
+static const char _rcsid[] = "$Id: _cdio_linux.c,v 1.27 2003/10/03 01:26:52 rocky Exp $";
 
 #include <string.h>
 
@@ -35,6 +35,7 @@ static const char _rcsid[] = "$Id: _cdio_linux.c,v 1.16 2003/09/01 15:11:36 rock
 #include <cdio/util.h>
 #include "cdio_assert.h"
 #include "cdio_private.h"
+#include "scsi_mmc.h"
 
 #ifdef HAVE_LINUX_CDROM
 
@@ -96,7 +97,7 @@ typedef struct {
    Return 1 if a CD-ROM. 0 if it exists but isn't a CD-ROM drive
    and -1 if no device exists .
 */
-static int
+static bool
 cdio_is_cdrom(char *drive, char *mnttype)
 {
   bool is_cd=false;
@@ -235,25 +236,19 @@ _set_bsize (int fd, unsigned int bsize)
    Can read only up to 25 blocks.
 */
 static int
-_read_packet_audio_sector (int fd, void *buf, lba_t lba)
+_cdio_mmc_read_sectors (int fd, void *buf, lba_t lba, int sector_type, 
+			unsigned int nblocks)
 {
-  struct cdrom_generic_command cgc;
-  const int nblocks = 1;
+  typedef struct cdrom_generic_command cgc_t;
+  cgc_t cgc;
 
-  memset (&cgc, 0, sizeof (struct cdrom_generic_command));
+  memset (&cgc, 0, sizeof (cgc_t));
 
-  cgc.cmd[0] = GPCMD_READ_CD;
-  cgc.cmd[1] = 0; /* Any type of sectors returned */
-  
-  cgc.cmd[2] = (lba >> 24) & 0xff;
-  cgc.cmd[3] = (lba >> 16) & 0xff;
-  cgc.cmd[4] = (lba >> 8) & 0xff;
-  cgc.cmd[5] = (lba >> 0) & 0xff;
-
-  cgc.cmd[6] = (nblocks >> 16) & 0xff;
-  cgc.cmd[7] = (nblocks >> 8) & 0xff;
-  cgc.cmd[8] = (nblocks >> 0) & 0xff;
-  cgc.cmd[9] = 0x78; /* All headers  */
+  cgc.cmd[0] = CDIO_MMC_GPCMD_READ_CD;
+  CDIO_MMC_SET_READ_TYPE  (cgc.cmd, sector_type);
+  CDIO_MMC_SET_READ_LBA   (cgc.cmd, lba);
+  CDIO_MMC_SET_READ_LENGTH(cgc.cmd, nblocks);
+  CDIO_MMC_SET_MAIN_CHANNEL_SELECTION_BITS(cgc.cmd, CDIO_MMC_MCSB_ALL_HEADERS);
 
   cgc.buflen = CDIO_CD_FRAMESIZE_RAW * nblocks; 
   cgc.buffer = buf;
@@ -268,6 +263,18 @@ _read_packet_audio_sector (int fd, void *buf, lba_t lba)
   return 0;
 }
 
+/* MMC driver to read audio sectors. 
+   Can read only up to 25 blocks.
+*/
+static int
+_cdio_read_audio_sectors (void *env, void *buf, lsn_t lsn, 
+			  unsigned int nblocks)
+{
+  _img_private_t *_obj = env;
+  return _cdio_mmc_read_sectors( _obj->gen.fd, buf, lsn, 
+				 CDIO_MMC_READ_TYPE_CDDA, nblocks);
+}
+
 /* Packet driver to read mode2 sectors. 
    Can read only up to 25 blocks.
 */
@@ -279,16 +286,10 @@ __read_packet_mode2_sectors (int fd, void *buf, lba_t lba,
 
   memset (&cgc, 0, sizeof (struct cdrom_generic_command));
 
-  cgc.cmd[0] = use_read_10 ? GPCMD_READ_10 : GPCMD_READ_CD;
-  
-  cgc.cmd[2] = (lba >> 24) & 0xff;
-  cgc.cmd[3] = (lba >> 16) & 0xff;
-  cgc.cmd[4] = (lba >> 8) & 0xff;
-  cgc.cmd[5] = (lba >> 0) & 0xff;
+  cgc.cmd[0] = use_read_10 ? GPCMD_READ_10 : CDIO_MMC_GPCMD_READ_CD;
 
-  cgc.cmd[6] = (nblocks >> 16) & 0xff;
-  cgc.cmd[7] = (nblocks >> 8) & 0xff;
-  cgc.cmd[8] = (nblocks >> 0) & 0xff;
+  CDIO_MMC_SET_READ_LBA(cgc.cmd, lba);
+  CDIO_MMC_SET_READ_LENGTH(cgc.cmd, nblocks);
 
   if (!use_read_10)
     {
@@ -357,14 +358,14 @@ _read_packet_mode2_sectors (int fd, void *buf, lba_t lba,
    from lsn. Returns 0 if no error. 
  */
 static int
-_cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn, 
+_cdio_read_mode2_sector (void *env, void *data, lsn_t lsn, 
 			 bool mode2_form2)
 {
   char buf[M2RAW_SECTOR_SIZE] = { 0, };
   struct cdrom_msf *msf = (struct cdrom_msf *) &buf;
   msf_t _msf;
 
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   cdio_lba_to_msf (cdio_lsn_to_lba(lsn), &_msf);
   msf->cdmsf_min0 = from_bcd8(_msf.m);
@@ -419,73 +420,17 @@ _cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn,
   return 0;
 }
 
-/*!
-   Reads a single audio sector from CD device into data starting
-   from lsn. Returns 0 if no error. 
- */
-static int
-_cdio_read_audio_sector (void *user_data, void *data, lsn_t lsn)
-{
-  char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
-  struct cdrom_msf *msf = (struct cdrom_msf *) &buf;
-  msf_t _msf;
-
-  _img_private_t *_obj = user_data;
-
-  cdio_lba_to_msf (cdio_lsn_to_lba(lsn), &_msf);
-  msf->cdmsf_min0 = from_bcd8(_msf.m);
-  msf->cdmsf_sec0 = from_bcd8(_msf.s);
-  msf->cdmsf_frame0 = from_bcd8(_msf.f);
-
- retry:
-  switch (_obj->access_mode)
-    {
-    case _AM_NONE:
-      cdio_error ("no way to read audio");
-      return 1;
-      break;
-
-    case _AM_IOCTL:
-      if (ioctl (_obj->gen.fd, CDROMREADRAW, &buf) == -1) {
-	perror ("ioctl()");
-	return 1;
-	/* exit (EXIT_FAILURE); */
-      }
-      break;
-
-    case _AM_READ_CD:
-    case _AM_READ_10:
-      if (_read_packet_audio_sector (_obj->gen.fd, buf, lsn)) {
-	perror ("ioctl()");
-	if (_obj->access_mode == _AM_READ_CD) {
-	  cdio_info ("READ_CD failed; switching to READ_10 mode...");
-	  _obj->access_mode = _AM_READ_10;
-	  goto retry;
-	} else {
-	  cdio_info ("READ_10 failed; switching to ioctl(CDROMREADAUDIO) mode...");
-	  _obj->access_mode = _AM_IOCTL;
-	  goto retry;
-	}
-	return 1;
-      }
-      break;
-    }
-
-  memcpy (data, buf, CDIO_CD_FRAMESIZE_RAW);
-  return 0;
-}
-
 /*!
    Reads nblocks of mode2 sectors from cd device into data starting
    from lsn.
    Returns 0 if no error. 
  */
 static int
-_cdio_read_mode2_sectors (void *user_data, void *data, lsn_t lsn, 
+_cdio_read_mode2_sectors (void *env, void *data, lsn_t lsn, 
 			  bool mode2_form2, unsigned int nblocks)
 {
-  _img_private_t *_obj = user_data;
-  int i;
+  _img_private_t *_obj = env;
+  unsigned int i;
   int retval;
 
   for (i = 0; i < nblocks; i++) {
@@ -511,9 +456,9 @@ _cdio_read_mode2_sectors (void *user_data, void *data, lsn_t lsn,
    Return the size of the CD in logical block address (LBA) units.
  */
 static uint32_t 
-_cdio_stat_size (void *user_data)
+_cdio_stat_size (void *env)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   struct cdrom_tocentry tocent;
   uint32_t size;
@@ -532,12 +477,17 @@ _cdio_stat_size (void *user_data)
 }
 
 /*!
-  Set the key "arg" to "value" in source device.
+  Set the arg "key" with "value" in the source device.
+  Currently "source" and "access-mode" are valid keys.
+  "source" sets the source device in I/O operations 
+  "access-mode" sets the the method of CD access 
+
+  0 is returned if no error was found, and nonzero if there as an error.
 */
 static int
-_cdio_set_arg (void *user_data, const char key[], const char value[])
+_cdio_set_arg (void *env, const char key[], const char value[])
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   if (!strcmp (key, "source"))
     {
@@ -556,8 +506,9 @@ _cdio_set_arg (void *user_data, const char key[], const char value[])
 	_obj->access_mode = _AM_READ_CD;
       else if (!strcmp(value, "READ_10"))
 	_obj->access_mode = _AM_READ_10;
-      else
-	cdio_error ("unknown access type: %s. ignored.", value);
+      else {
+	cdio_warn ("unknown access type: %s. ignored.", value);
+      }
     }
   else 
     return -1;
@@ -682,9 +633,9 @@ _cdio_eject_scsi(int fd)
   Return 0 if success and 1 for failure, and 2 if no routine.
  */
 static int 
-_cdio_eject_media (void *user_data) {
+_cdio_eject_media (void *env) {
 
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   int ret=2;
   int status;
   int fd;
@@ -729,9 +680,9 @@ _cdio_eject_media (void *user_data) {
   Return the value associated with the key "arg".
 */
 static const char *
-_cdio_get_arg (void *user_data, const char key[])
+_cdio_get_arg (void *env, const char key[])
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   if (!strcmp (key, "source")) {
     return _obj->gen.source_name;
@@ -755,23 +706,40 @@ _cdio_get_arg (void *user_data, const char key[])
   CDIO_INVALID_TRACK is returned on error.
 */
 static track_t
-_cdio_get_first_track_num(void *user_data) 
+_cdio_get_first_track_num(void *env) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   if (!_obj->gen.toc_init) _cdio_read_toc (_obj) ;
 
   return FIRST_TRACK_NUM;
 }
 
+/*!
+  Return the media catalog number MCN.
+
+  Note: string is malloc'd so caller should free() then returned
+  string when done with it.
+
+ */
+static char *
+_cdio_get_mcn (void *env) {
+
+  struct cdrom_mcn mcn;
+  _img_private_t *_obj = env;
+  if (ioctl(_obj->gen.fd, CDROM_GET_MCN, &mcn) != 0)
+    return NULL;
+  return strdup(mcn.medium_catalog_number);
+}
+
 /*!
   Return the number of tracks in the current medium.
   CDIO_INVALID_TRACK is returned on error.
 */
 static track_t
-_cdio_get_num_tracks(void *user_data) 
+_cdio_get_num_tracks(void *env) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   if (!_obj->gen.toc_init) _cdio_read_toc (_obj) ;
 
@@ -782,9 +750,9 @@ _cdio_get_num_tracks(void *user_data)
   Get format of track. 
 */
 static track_format_t
-_cdio_get_track_format(void *user_data, track_t track_num) 
+_cdio_get_track_format(void *env, track_t track_num) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   if (!_obj->gen.toc_init) _cdio_read_toc (_obj) ;
 
@@ -815,9 +783,9 @@ _cdio_get_track_format(void *user_data, track_t track_num)
   FIXME: there's gotta be a better design for this and get_track_format?
 */
 static bool
-_cdio_get_track_green(void *user_data, track_t track_num) 
+_cdio_get_track_green(void *env, track_t track_num) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   if (!_obj->gen.toc_init) _cdio_read_toc (_obj) ;
 
@@ -840,9 +808,9 @@ _cdio_get_track_green(void *user_data, track_t track_num)
   False is returned if there is no track entry.
 */
 static bool
-_cdio_get_track_msf(void *user_data, track_t track_num, msf_t *msf)
+_cdio_get_track_msf(void *env, track_t track_num, msf_t *msf)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   if (NULL == msf) return false;
 
@@ -861,10 +829,76 @@ _cdio_get_track_msf(void *user_data, track_t track_num, msf_t *msf)
   }
 }
 
+/* checklist: /dev/cdrom, /dev/dvd /dev/hd?, /dev/scd? /dev/sr? */
+static char checklist1[][40] = {
+  {"cdrom"}, {"dvd"}, {""}
+};
+static char checklist2[][40] = {
+  {"?a hd?"}, {"?0 scd?"}, {"?0 sr?"}, {""}
+};
+
 #endif /* HAVE_LINUX_CDROM */
 
 /*!
-  Return a string containing the default VCD device if none is specified.
+  Return an array of strings giving possible CD devices.
+ */
+char **
+cdio_get_devices_linux (void)
+{
+#ifndef HAVE_LINUX_CDROM
+  return NULL;
+#else
+  unsigned int i;
+  char drive[40];
+  char *ret_drive;
+  bool exists;
+  char **drives = NULL;
+  unsigned int num_drives=0;
+  
+  /* Scan the system for CD-ROM drives.
+  */
+  for ( i=0; strlen(checklist1[i]) > 0; ++i ) {
+    sprintf(drive, "/dev/%s", checklist1[i]);
+    if ( (exists=cdio_is_cdrom(drive, NULL)) > 0 ) {
+      cdio_add_device_list(&drives, drive, &num_drives);
+    }
+  }
+
+  /* Now check the currently mounted CD drives */
+  if (NULL != (ret_drive = cdio_check_mounts("/etc/mtab"))) {
+    cdio_add_device_list(&drives, drive, &num_drives);
+  }
+  
+  /* Finally check possible mountable drives in /etc/fstab */
+  if (NULL != (ret_drive = cdio_check_mounts("/etc/fstab"))) {
+    cdio_add_device_list(&drives, drive, &num_drives);
+  }
+
+  /* Scan the system for CD-ROM drives.
+     Not always 100% reliable, so use the USE_MNTENT code above first.
+  */
+  for ( i=0; strlen(checklist2[i]) > 0; ++i ) {
+    unsigned int j;
+    char *insert;
+    exists = true;
+    for ( j=checklist2[i][1]; exists; ++j ) {
+      sprintf(drive, "/dev/%s", &checklist2[i][3]);
+      insert = strchr(drive, '?');
+      if ( insert != NULL ) {
+	*insert = j;
+      }
+      if ( (exists=cdio_is_cdrom(drive, NULL)) > 0 ) {
+	cdio_add_device_list(&drives, drive, &num_drives);
+      }
+    }
+  }
+  cdio_add_device_list(&drives, NULL, &num_drives);
+  return drives;
+#endif /*HAVE_LINUX_CDROM*/
+}
+
+/*!
+  Return a string containing the default CD device.
  */
 char *
 cdio_get_default_device_linux(void)
@@ -873,16 +907,9 @@ cdio_get_default_device_linux(void)
   return NULL;
   
 #else
-  /* checklist: /dev/cdrom, /dev/dvd /dev/hd?, /dev/scd? /dev/sr? */
-  static char checklist1[][40] = {
-    {"cdrom"}, {"dvd"}, {""}
-  };
-  static char checklist2[][40] = {
-    {"?a hd?"}, {"?0 scd?"}, {"?0 sr?"}, {""}
-  };
   unsigned int i;
   char drive[40];
-  int exists;
+  bool exists;
   char *ret_drive;
 
   /* Scan the system for CD-ROM drives.
@@ -908,7 +935,7 @@ cdio_get_default_device_linux(void)
   for ( i=0; strlen(checklist2[i]) > 0; ++i ) {
     unsigned int j;
     char *insert;
-    exists = 1;
+    exists = true;
     for ( j=checklist2[i][1]; exists; ++j ) {
       sprintf(drive, "/dev/%s", &checklist2[i][3]);
       insert = strchr(drive, '?');
@@ -941,8 +968,10 @@ cdio_open_linux (const char *orig_source_name)
     .eject_media        = _cdio_eject_media,
     .free               = cdio_generic_free,
     .get_arg            = _cdio_get_arg,
+    .get_devices        = cdio_get_devices_linux,
     .get_default_device = cdio_get_default_device_linux,
     .get_first_track_num= _cdio_get_first_track_num,
+    .get_mcn            = _cdio_get_mcn,
     .get_num_tracks     = _cdio_get_num_tracks,
     .get_track_format   = _cdio_get_track_format,
     .get_track_green    = _cdio_get_track_green,
@@ -950,7 +979,7 @@ cdio_open_linux (const char *orig_source_name)
     .get_track_msf      = _cdio_get_track_msf,
     .lseek              = cdio_generic_lseek,
     .read               = cdio_generic_read,
-    .read_audio_sector  = _cdio_read_audio_sector,
+    .read_audio_sectors = _cdio_read_audio_sectors,
     .read_mode2_sector  = _cdio_read_mode2_sector,
     .read_mode2_sectors = _cdio_read_mode2_sectors,
     .set_arg            = _cdio_set_arg,
