@@ -1,5 +1,5 @@
 /*
-    $Id: _cdio_nrg.c,v 1.15 2003/07/27 22:52:22 rocky Exp $
+    $Id: _cdio_nrg.c,v 1.22 2003/10/07 03:11:38 rocky Exp $
 
     Copyright (C) 2001,2003 Herbert Valerio Riedel <hvr@gnu.org>
 
@@ -25,9 +25,18 @@
 # include "config.h"
 #endif
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_GLOB_H
+#include <glob.h>
+#endif
 
 #include <cdio/logging.h>
 #include <cdio/sector.h>
@@ -38,7 +47,7 @@
 #include "cdio_private.h"
 #include "_cdio_stdio.h"
 
-static const char _rcsid[] = "$Id: _cdio_nrg.c,v 1.15 2003/07/27 22:52:22 rocky Exp $";
+static const char _rcsid[] = "$Id: _cdio_nrg.c,v 1.22 2003/10/07 03:11:38 rocky Exp $";
 
 /* structures used */
 
@@ -92,6 +101,9 @@ PRAGMA_END_PACKED
 #define NER5_ID  0x4e455235
 #define NERO_ID  0x4e45524f
 #define SINF_ID  0x53494e46
+#define MTYP_ID  0x4d545950  /* Media type? */
+
+#define MTYP_AUDIO_CD 1
 
 /* reader */
 
@@ -139,11 +151,12 @@ typedef struct {
   track_t         total_tracks;    /* number of tracks in image */
   track_t         first_track_num; /* track number of first track */
   CdioList       *mapping;         /* List of track information */
-  uint32_t size;
+  uint32_t        size;
+  uint32_t        mtyp;            /* Value of MTYP tag */
 } _img_private_t;
 
 static bool     _cdio_parse_nero_footer (_img_private_t *_obj);
-static uint32_t _cdio_stat_size (void *user_data);
+static uint32_t _cdio_stat_size (void *env);
 
 /* Updates internal track TOC, so we can later 
    simulate ioctl(CDROMREADTOCENTRY).
@@ -210,6 +223,7 @@ _cdio_parse_nero_footer (_img_private_t *_obj)
     return 0;
 
   size = cdio_stream_stat (_obj->gen.data_source);
+  if (-1 == size) return false;
 
   {
 PRAGMA_BEGIN_PACKED
@@ -228,7 +242,7 @@ PRAGMA_END_PACKED
 
     cdio_assert (sizeof (buf) == 12);
  
- cdio_stream_seek (_obj->gen.data_source, size - sizeof (buf), SEEK_SET);
+    cdio_stream_seek (_obj->gen.data_source, size - sizeof (buf), SEEK_SET);
     cdio_stream_read (_obj->gen.data_source, (void *) &buf, sizeof (buf), 1);
     
     if (buf.v50.ID == UINT32_TO_BE (0x4e45524f)) /* "NERO" */
@@ -243,7 +257,7 @@ PRAGMA_END_PACKED
       }
     else
       {
-	cdio_error ("Image not recognized as either v50 or v55 type NRG");
+	cdio_warn ("Image not recognized as either v50 or v55 type NRG");
 	return -1;
       }
 
@@ -457,12 +471,27 @@ PRAGMA_END_PACKED
 	
 	cdio_assert (UINT32_FROM_BE (chunk->len) == 4);
 	
-	cdio_debug ("SINF: %d sessions", UINT32_FROM_BE (*_sessions));
+	cdio_debug ("SINF: %lu sessions", 
+		    (long unsigned int) UINT32_FROM_BE (*_sessions));
+      }
+	break;
+	
+      case MTYP_ID: { /* "MTYP" */
+	uint32_t *mtyp_p = (void *) chunk->data;
+	uint32_t mtyp  = UINT32_FROM_BE (*mtyp_p);
+	
+	cdio_assert (UINT32_FROM_BE (chunk->len) == 4);
+
+	if (mtyp != MTYP_AUDIO_CD) {
+	  cdio_warn ("Unknown media type: %u", (unsigned int) mtyp);
+	}
+	_obj->mtyp = mtyp;
       }
 	break;
 	
       default:
-	cdio_warn ("unknown tag %8.8x seen", UINT32_FROM_BE (chunk->id));
+	cdio_warn ("unknown tag %8.8x seen", 
+		   (unsigned int) UINT32_FROM_BE (chunk->id));
 	break;
       }
 	
@@ -515,9 +544,9 @@ _cdio_init (_img_private_t *_obj)
   information in each sector.
 */
 static off_t
-_cdio_lseek (void *user_data, off_t offset, int whence)
+_cdio_lseek (void *env, off_t offset, int whence)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   /* real_offset is the real byte offset inside the disk image
      The number below was determined empirically. I'm guessing
@@ -574,16 +603,16 @@ _cdio_lseek (void *user_data, off_t offset, int whence)
    boundaries.
 */
 static ssize_t
-_cdio_read (void *user_data, void *buf, size_t size)
+_cdio_read (void *env, void *buf, size_t size)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   return cdio_stream_read(_obj->gen.data_source, buf, size, 1);
 }
 
 static uint32_t 
-_cdio_stat_size (void *user_data)
+_cdio_stat_size (void *env)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   return _obj->size;
 }
@@ -593,16 +622,17 @@ _cdio_stat_size (void *user_data)
    from LSN. Returns 0 if no error. 
  */
 static int
-_cdio_read_audio_sector (void *user_data, void *data, lsn_t lsn)
+_cdio_read_audio_sectors (void *env, void *data, lsn_t lsn, 
+			  unsigned int nblocks)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   CdioListNode *node;
 
   if (lsn >= _obj->size)
     {
-      cdio_warn ("trying to read beyond image size (%d >= %u)", lsn, 
-		_obj->size);
+      cdio_warn ("trying to read beyond image size (%lu >= %lu)", 
+		 (long unsigned int) lsn, (long unsigned int) _obj->size);
       return -1;
     }
 
@@ -619,22 +649,23 @@ _cdio_read_audio_sector (void *user_data, void *data, lsn_t lsn)
 			      SEEK_SET); 
       if (ret!=0) return ret;
       ret = cdio_stream_read (_obj->gen.data_source, data, 
-			      CDIO_CD_FRAMESIZE_RAW, 1);
+			      CDIO_CD_FRAMESIZE_RAW, nblocks);
       if (ret==0) return ret;
       break;
     }
   }
 
-  if (!node) cdio_warn ("reading into pre gap (lsn %d)", lsn);
+  if (!node) cdio_warn ("reading into pre gap (lsn %lu)", 
+			(long unsigned int) lsn);
 
   return 0;
 }
 
 static int
-_cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn, 
+_cdio_read_mode2_sector (void *env, void *data, lsn_t lsn, 
 			 bool mode2_form2)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
   int blocksize = _obj->sector_2336 
     ? M2RAW_SECTOR_SIZE : CDIO_CD_FRAMESIZE_RAW;
@@ -643,8 +674,8 @@ _cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn,
 
   if (lsn >= _obj->size)
     {
-      cdio_warn ("trying to read beyond image size (%d >= %u)", lsn, 
-		_obj->size);
+      cdio_warn ("trying to read beyond image size (%lu >= %lu)", 
+		 (long unsigned int) lsn, (long unsigned int) _obj->size);
       return -1;
     }
 
@@ -670,7 +701,7 @@ _cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn,
   }
 
   if (!node)
-    cdio_warn ("reading into pre gap (lsn %d)", lsn);
+    cdio_warn ("reading into pre gap (lsn %lu)", (long unsigned int) lsn);
 
   if (mode2_form2)
     memcpy (data, buf + CDIO_CD_SYNC_SIZE + CDIO_CD_HEADER_SIZE, 
@@ -687,10 +718,10 @@ _cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn,
    Returns 0 if no error. 
  */
 static int
-_cdio_read_mode2_sectors (void *user_data, void *data, uint32_t lsn, 
+_cdio_read_mode2_sectors (void *env, void *data, uint32_t lsn, 
 		     bool mode2_form2, unsigned nblocks)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   int i;
   int retval;
 
@@ -716,9 +747,9 @@ _cdio_read_mode2_sectors (void *user_data, void *data, uint32_t lsn,
   Set the device to use in I/O operations.
 */
 static int
-_cdio_set_arg (void *user_data, const char key[], const char value[])
+_cdio_set_arg (void *env, const char key[], const char value[])
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   if (!strcmp (key, "source"))
     {
@@ -739,9 +770,9 @@ _cdio_set_arg (void *user_data, const char key[], const char value[])
   Return the value associated with the key "arg".
 */
 static const char *
-_cdio_get_arg (void *user_data, const char key[])
+_cdio_get_arg (void *env, const char key[])
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   if (!strcmp (key, "source")) {
     return _obj->gen.source_name;
@@ -750,12 +781,39 @@ _cdio_get_arg (void *user_data, const char key[])
 }
 
 /*!
-  Return a string containing the default VCD device if none is specified.
+  Return an array of strings giving possible NRG disk images.
+ */
+char **
+cdio_get_devices_nrg (void)
+{
+  char **drives = NULL;
+  unsigned int num_files=0;
+#ifdef HAVE_GLOB_H
+  unsigned int i;
+  glob_t globbuf;
+  globbuf.gl_offs = 0;
+  glob("*.nrg", GLOB_DOOFFS, NULL, &globbuf);
+  for (i=0; i<globbuf.gl_pathc; i++) {
+    cdio_add_device_list(&drives, globbuf.gl_pathv[i], &num_files);
+  }
+  globfree(&globbuf);
+#else
+  cdio_add_device_list(&drives, DEFAULT_CDIO_DEVICE, &num_files);
+#endif /*HAVE_GLOB_H*/
+  cdio_add_device_list(&drives, NULL, &num_files);
+  return drives;
+}
+
+/*!
+  Return a string containing the default CD device.
  */
 char *
-cdio_get_default_device_nrg()
+cdio_get_default_device_nrg(void)
 {
-  return strdup(DEFAULT_CDIO_DEVICE);
+  char **drives = cdio_get_devices_nrg();
+  char *drive = (drives[0] == NULL) ? NULL : strdup(drives[0]);
+  cdio_free_device_list(drives);
+  return drive;
 }
 
 /*!
@@ -763,9 +821,9 @@ cdio_get_default_device_nrg()
   CDIO_INVALID_TRACK is returned on error.
 */
 static track_t
-_cdio_get_first_track_num(void *user_data) 
+_cdio_get_first_track_num(void *env) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   return _obj->first_track_num;
 }
@@ -775,9 +833,9 @@ _cdio_get_first_track_num(void *user_data)
   one big track. 
 */
 static track_t
-_cdio_get_num_tracks(void *user_data) 
+_cdio_get_num_tracks(void *env) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   return _obj->total_tracks;
 }
@@ -787,13 +845,14 @@ _cdio_get_num_tracks(void *user_data)
   CDIO_INVALID_TRACK is returned on error.
 */
 static track_format_t
-_cdio_get_track_format(void *user_data, track_t track_num) 
+_cdio_get_track_format(void *env, track_t track_num) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   if (track_num > _obj->total_tracks || track_num == 0) 
     return TRACK_FORMAT_ERROR;
 
+  if ( MTYP_AUDIO_CD == _obj->mtyp) return TRACK_FORMAT_AUDIO;
   return _obj->tocent[track_num-1].track_format;
 }
 
@@ -806,13 +865,14 @@ _cdio_get_track_format(void *user_data, track_t track_num)
   FIXME: there's gotta be a better design for this and get_track_format?
 */
 static bool
-_cdio_get_track_green(void *user_data, track_t track_num) 
+_cdio_get_track_green(void *env, track_t track_num) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
   
   if (track_num > _obj->total_tracks || track_num == 0) 
     return false;
 
+  if ( MTYP_AUDIO_CD == _obj->mtyp) return false;
   return _obj->tocent[track_num-1].track_green;
 }
 
@@ -823,9 +883,9 @@ _cdio_get_track_green(void *user_data, track_t track_num)
   using track_num LEADOUT_TRACK or the total tracks+1.
 */
 static bool
-_cdio_get_track_msf(void *user_data, track_t track_num, msf_t *msf)
+_cdio_get_track_msf(void *env, track_t track_num, msf_t *msf)
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
 
   if (NULL == msf) return 1;
 
@@ -848,6 +908,7 @@ cdio_open_nrg (const char *source_name)
     .eject_media        = cdio_generic_bogus_eject_media,
     .free               = cdio_generic_stream_free,
     .get_arg            = _cdio_get_arg,
+    .get_devices        = cdio_get_devices_nrg,
     .get_default_device = cdio_get_default_device_nrg,
     .get_first_track_num= _cdio_get_first_track_num,
     .get_num_tracks     = _cdio_get_num_tracks,
@@ -857,7 +918,7 @@ cdio_open_nrg (const char *source_name)
     .get_track_msf      = _cdio_get_track_msf,
     .lseek              = _cdio_lseek,
     .read               = _cdio_read,
-    .read_audio_sector  = _cdio_read_audio_sector,
+    .read_audio_sectors = _cdio_read_audio_sectors,
     .read_mode2_sector  = _cdio_read_mode2_sector,
     .read_mode2_sectors = _cdio_read_mode2_sectors,
     .set_arg            = _cdio_set_arg,
@@ -869,6 +930,7 @@ cdio_open_nrg (const char *source_name)
 
 
   _data->total_tracks   = 0;
+  _data->mtyp           = 0; 
   _data->first_track_num= 1;
   _data->sector_2336    = false; /* FIXME: remove sector_2336 */
   _data->is_cues        = false; /* FIXME: remove is_cues. */
