@@ -1,5 +1,5 @@
 /*
-    $Id: _cdio_win32.c,v 1.10 2003/08/11 09:19:18 rocky Exp $
+    $Id: _cdio_win32.c,v 1.17 2003/10/17 02:25:36 rocky Exp $
 
     Copyright (C) 2003 Rocky Bernstein <rocky@panix.com>
 
@@ -26,13 +26,14 @@
 # include "config.h"
 #endif
 
-static const char _rcsid[] = "$Id: _cdio_win32.c,v 1.10 2003/08/11 09:19:18 rocky Exp $";
+static const char _rcsid[] = "$Id: _cdio_win32.c,v 1.17 2003/10/17 02:25:36 rocky Exp $";
 
 #include <cdio/cdio.h>
 #include <cdio/sector.h>
 #include <cdio/util.h>
 #include "cdio_assert.h"
 #include "cdio_private.h"
+#include "scsi_mmc.h"
 
 /* LBA = msf.frame + 75 * ( msf.second - 2 + 60 * msf.minute ) */
 #define MSF_TO_LBA2(min, sec, frame) ((int)frame + 75 * (sec -2 + 60 * min))
@@ -59,7 +60,7 @@ static const char _rcsid[] = "$Id: _cdio_win32.c,v 1.10 2003/08/11 09:19:18 rock
 #    define MAXIMUM_NUMBER_TRACKS 100
 #endif
 typedef struct _TRACK_DATA {
-    UCHAR Reserved;
+    UCHAR Format;
     UCHAR Control : 4;
     UCHAR Adr : 4;
     UCHAR TrackNumber;
@@ -112,7 +113,6 @@ typedef struct __RAW_READ_INFO {
 #define SRB_DIR_OUT         0x10
 #define SRB_EVENT_NOTIFY    0x40
 
-#define READ_CD 0xbe
 #define SECTOR_TYPE_MODE2 0x14
 #define READ_CD_USERDATA_MODE2 0x10
 
@@ -176,6 +176,8 @@ struct SRB_ExecSCSICmd
 
 typedef struct {
   lsn_t          start_lsn;
+  UCHAR          Control : 4;
+  UCHAR          Format;
 } track_info_t;
 
 typedef struct {
@@ -485,18 +487,25 @@ _cdio_win32_free (void *user_data)
    Returns 0 if no error. 
  */
 static int
-_cdio_read_raw_sector (void *user_data, void *data, lsn_t lsn)
+_cdio_mmc_read_sectors (void *user_data, void *data, lsn_t lsn, 
+			int sector_type, unsigned int nblocks)
 {
-  unsigned char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
   _img_private_t *_obj = user_data;
+  unsigned char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
 
   if( _obj->hASPI ) {
     HANDLE hEvent;
     struct SRB_ExecSCSICmd ssc;
-    int blocks = 1;
-    int sector_type;
-    int sync, header_code, user_data, edc_ecc, error_field;
-    int sub_channel;
+
+#if 1
+    int sector_type = 0; /*all types */
+    int sync        = 0;
+    int header_code = 2;
+    int user_data   = 1;
+    int edc_ecc     = 0;
+    int error_field = 0;
+#endif
+	
     
     /* Create the transfer completion event */
     hEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
@@ -505,18 +514,6 @@ _cdio_read_raw_sector (void *user_data, void *data, lsn_t lsn)
     }
 
     /* Data selection */
-    sector_type = 0;		/* all types */
-    /*sector_type = 1;*/	/* CD-DA */
-    /*sector_type = 2;*/	/* mode1 */
-    /*sector_type = 3;*/	/* mode2 */
-    /*sector_type = 4;*/	/* mode2/form1 */
-    /*sector_type = 5;*/	/* mode2/form2 */
-    sync = 0;
-    header_code = 2;
-    user_data = 1;
-    edc_ecc = 0;
-    error_field = 0;
-    sub_channel = 0;
 
     memset( &ssc, 0, sizeof( ssc ) );
     
@@ -530,26 +527,23 @@ _cdio_read_raw_sector (void *user_data, void *data, lsn_t lsn)
     ssc.SRB_CDBLen      = 12;
     
     /* Operation code */
-    ssc.CDBByte[ 0 ] = READ_CD;
+    ssc.CDBByte[ 0 ] = CDIO_MMC_GPCMD_READ_CD;
 
-    /* Start of LSN */
-    ssc.CDBByte[ 1 ] = (sector_type) << 2;
-    ssc.CDBByte[ 2 ] = (lsn >> 24) & 0xff;
-    ssc.CDBByte[ 3 ] = (lsn >> 16) & 0xff;
-    ssc.CDBByte[ 4 ] = (lsn >>  8) & 0xff;
-    ssc.CDBByte[ 5 ] = (lsn      ) & 0xff;
-    
-    /* Transfer length */
-    ssc.CDBByte[ 6 ] = (blocks >> 16) & 0xff;
-    ssc.CDBByte[ 7 ] = (blocks >>  8) & 0xff;
-    ssc.CDBByte[ 8 ] = blocks & 0xff;
-    
+    CDIO_MMC_SET_READ_TYPE(ssc.CDBByte, sector_type);
+    CDIO_MMC_SET_READ_LBA(ssc.CDBByte, lsn);
+    CDIO_MMC_SET_READ_LENGTH(ssc.CDBByte, nblocks);
+
+#if 1
     ssc.CDBByte[ 9 ] = (sync << 7) |
       (header_code << 5) |
       (user_data << 4) |
       (edc_ecc << 3) |
       (error_field << 1);
     /* ssc.CDBByte[ 9 ] = READ_CD_USERDATA_MODE2; */
+#else 
+    CDIO_MMC_SET_MAIN_CHANNEL_SELECTION_BITS(ssc.CDBByte,
+					     CDIO_MMC_MCSB_ALL_HEADERS);
+#endif
     
     /* Result buffer */
     ssc.SRB_BufPointer  = buf;
@@ -595,6 +589,18 @@ _cdio_read_raw_sector (void *user_data, void *data, lsn_t lsn)
   return 0;
 }
 
+/*!
+   Reads an audio device into data starting from lsn.
+   Returns 0 if no error. 
+ */
+static int
+_cdio_read_audio_sectors (void *user_data, void *data, lsn_t lsn, 
+			  unsigned int nblocks) 
+{
+  return _cdio_mmc_read_sectors( user_data, data, lsn, 
+				 CDIO_MMC_READ_TYPE_CDDA, nblocks );
+}
+
 /*!
    Reads a single mode2 sector from cd device into data starting
    from lsn. Returns 0 if no error. 
@@ -617,12 +623,33 @@ _cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn,
       || (_obj->gen.ioctls_debugged < (30 * 75)  
 	  && _obj->gen.ioctls_debugged % 75 == 0)
       || _obj->gen.ioctls_debugged % (30 * 75) == 0)
-    cdio_debug ("reading %d", lsn);
+    cdio_debug ("reading %lu", (unsigned long int) lsn);
   
   _obj->gen.ioctls_debugged++;
- 
-  ret = _cdio_read_raw_sector(user_data, buf, lsn);
 
+  if ( _obj->hASPI ) {
+    ret = _cdio_mmc_read_sectors(user_data, buf, lsn, 
+				 CDIO_MMC_READ_TYPE_ANY, 1);
+  } else {
+    DWORD dwBytesReturned;
+    RAW_READ_INFO cdrom_raw;
+
+    /* Initialize CDROM_RAW_READ structure */
+    cdrom_raw.DiskOffset.QuadPart = CDIO_CD_FRAMESIZE * lsn;
+    cdrom_raw.SectorCount = 1;
+    cdrom_raw.TrackMode = XAForm2;
+    
+    if( DeviceIoControl( _obj->h_device_handle,
+			 IOCTL_CDROM_RAW_READ, &cdrom_raw,
+			 sizeof(RAW_READ_INFO), buf,
+			 sizeof(buf), &dwBytesReturned, NULL )
+	== 0 ) {
+      cdio_debug("Error reading %lu\n", lsn);
+      return 1;
+      }
+    ret = 0;
+  }
+    
   if( ret != 0 ) return ret;
   
   if (mode2_form2)
@@ -640,7 +667,7 @@ _cdio_read_mode2_sector (void *user_data, void *data, lsn_t lsn,
  */
 static int
 _cdio_read_mode2_sectors (void *user_data, void *data, lsn_t lsn, 
-		     bool mode2_form2, unsigned nblocks)
+		     bool mode2_form2, unsigned int nblocks)
 {
   _img_private_t *_obj = user_data;
   int i;
@@ -801,14 +828,17 @@ _cdio_read_toc (_img_private_t *_obj)
 	  ((int)p_fulltoc[ i_index+1 ] << 16) +
 	  ((int)p_fulltoc[ i_index+2 ] << 8) +
 	  (int)p_fulltoc[ i_index+3 ];
+	_obj->tocent[ i ].Control   = (UCHAR)p_fulltoc[ 1 + 8 * i ];
 	
-	cdio_debug( "p_sectors: %i %i", i, _obj->tocent[i].start_lsn );
+	cdio_debug( "p_sectors: %i %lu", 
+		    i, (unsigned long int) _obj->tocent[i].start_lsn );
       }
 	
       free( p_fulltoc );
     }
     
     CloseHandle( hEvent );
+    _obj->gen.toc_init = true;
     return true;
     
   } else {
@@ -833,9 +863,13 @@ _cdio_read_toc (_img_private_t *_obj)
 					 cdrom_toc.TrackData[i].Address[1],
 					 cdrom_toc.TrackData[i].Address[2],
 					 cdrom_toc.TrackData[i].Address[3] );
-	cdio_debug("p_sectors: %i, %i", i, (_obj->tocent[i].start_lsn));
+	_obj->tocent[ i ].Control   = cdrom_toc.TrackData[i].Control;
+	_obj->tocent[ i ].Format    = cdrom_toc.TrackData[i].Format;
+	cdio_debug("p_sectors: %i, %lu", i, 
+		   (unsigned long int) (_obj->tocent[i].start_lsn));
       }
   }
+  _obj->gen.toc_init = true;
   return true;
 }
 
@@ -929,43 +963,61 @@ _cdio_get_num_tracks(void *user_data)
   Get format of track. 
 */
 static track_format_t
-_cdio_get_track_format(void *user_data, track_t track_num) 
+_cdio_get_track_format(void *env, track_t track_num) 
 {
-  _img_private_t *_obj = user_data;
+  _img_private_t *_obj = env;
+  
+  if (!_obj->gen.toc_init) _cdio_read_toc (_obj) ;
+
+  if (track_num > _obj->total_tracks || track_num == 0)
+    return TRACK_FORMAT_ERROR;
 
   MCI_OPEN_PARMS op;
   MCI_STATUS_PARMS st;
   DWORD i_flags;
-  char psz_drive[4];
   int ret;
-  unsigned int len = strlen(_obj->gen.source_name);
 
-  memset( &op, 0, sizeof(MCI_OPEN_PARMS) );
-  op.lpstrDeviceType = (LPCSTR)MCI_DEVTYPE_CD_AUDIO;
-  op.lpstrElementName = _obj->gen.source_name;
-  
-  /* Set the flags for the device type */
-  i_flags = MCI_OPEN_TYPE | MCI_OPEN_TYPE_ID |
-    MCI_OPEN_ELEMENT | MCI_OPEN_SHAREABLE;
-  
-  if( !mciSendCommand( 0, MCI_OPEN, i_flags, (unsigned long)&op ) ) {
-    st.dwItem  = MCI_CDA_STATUS_TYPE_TRACK;
-    st.dwTrack = track_num;
-    i_flags = MCI_TRACK | MCI_STATUS_ITEM ;
-    ret = mciSendCommand( op.wDeviceID, MCI_STATUS, i_flags, 
-			  (unsigned long) &st );
-
-    /* Release access to the device */
-    mciSendCommand( op.wDeviceID, MCI_CLOSE, MCI_WAIT, 0 );
-
-    switch(st.dwReturn) {
-    case MCI_CDA_TRACK_AUDIO:
-      return TRACK_FORMAT_AUDIO;
-    case MCI_CDA_TRACK_OTHER:
-      return TRACK_FORMAT_DATA;
-    default:
-      return TRACK_FORMAT_XA;
+  if( _obj->hASPI ) {
+    memset( &op, 0, sizeof(MCI_OPEN_PARMS) );
+    op.lpstrDeviceType = (LPCSTR)MCI_DEVTYPE_CD_AUDIO;
+    op.lpstrElementName = _obj->gen.source_name;
+    
+    /* Set the flags for the device type */
+    i_flags = MCI_OPEN_TYPE | MCI_OPEN_TYPE_ID |
+      MCI_OPEN_ELEMENT | MCI_OPEN_SHAREABLE;
+    
+    if( !_cdio_mciSendCommand( 0, MCI_OPEN, i_flags, &op ) ) {
+      st.dwItem  = MCI_CDA_STATUS_TYPE_TRACK;
+      st.dwTrack = track_num;
+      i_flags = MCI_TRACK | MCI_STATUS_ITEM ;
+      ret = mciSendCommand( op.wDeviceID, MCI_STATUS, i_flags, 
+			    (unsigned long) &st );
+      
+      /* Release access to the device */
+      mciSendCommand( op.wDeviceID, MCI_CLOSE, MCI_WAIT, 0 );
+      
+      switch(st.dwReturn) {
+      case MCI_CDA_TRACK_AUDIO:
+	return TRACK_FORMAT_AUDIO;
+      case MCI_CDA_TRACK_OTHER:
+	return TRACK_FORMAT_DATA;
+      default:
+	return TRACK_FORMAT_XA;
+      }
     }
+  } else {
+    /* This is pretty much copied from the "badly broken" cdrom_count_tracks
+       in linux/cdrom.c.
+    */
+    if (_obj->tocent[track_num-1].Control & 0x04) {
+      if (_obj->tocent[track_num-1].Format == 0x10)
+	return TRACK_FORMAT_CDI;
+      else if (_obj->tocent[track_num-1].Format == 0x20) 
+	return TRACK_FORMAT_XA;
+      else
+	return TRACK_FORMAT_DATA;
+    } else
+      return TRACK_FORMAT_AUDIO;
   }
 
   return TRACK_FORMAT_ERROR;
@@ -992,8 +1044,10 @@ _cdio_get_track_green(void *user_data, track_t track_num)
   if (track_num > _obj->total_tracks+1 || track_num == 0)
     return false;
 
-  /* FIXME! */
-  return true;
+  /* FIXME: Dunno if this is the right way, but it's what 
+     I was using in cd-info for a while.
+   */
+  return ((_obj->tocent[track_num-1].Control & 2) != 0);
 }
 
 /*!  
@@ -1024,6 +1078,48 @@ _cdio_get_track_msf(void *user_data, track_t track_num, msf_t *msf)
 
 #endif /* HAVE_WIN32_CDROM */
 
+/*!
+  Return an array of strings giving possible CD devices.
+ */
+char **
+cdio_get_devices_win32 (void)
+{
+#ifndef HAVE_WIN32_CDROM
+  return NULL;
+#else
+  char **drives = NULL;
+  unsigned int num_drives=0;
+  char drive_letter;
+  
+  /* Scan the system for CD-ROM drives.
+  */
+
+#if FINISHED
+  /* Now check the currently mounted CD drives */
+  if (NULL != (ret_drive = cdio_check_mounts("/etc/mtab"))) {
+    cdio_add_device_list(&drives, drive, &num_drives);
+  }
+  
+  /* Finally check possible mountable drives in /etc/fstab */
+  if (NULL != (ret_drive = cdio_check_mounts("/etc/fstab"))) {
+    cdio_add_device_list(&drives, drive, &num_drives);
+  }
+#endif
+
+  /* Scan the system for CD-ROM drives.
+     Not always 100% reliable, so use the USE_MNTENT code above first.
+  */
+  for (drive_letter='A'; drive_letter <= 'Z'; drive_letter++) {
+    const char *drive_str=cdio_is_cdrom(drive_letter);
+    if (drive_str != NULL) {
+      cdio_add_device_list(&drives, drive_str, &num_drives);
+    }
+  }
+  cdio_add_device_list(&drives, NULL, &num_drives);
+  return drives;
+#endif /*HAVE_WIN32_CDROM*/
+}
+
 /*!
   Return a string containing the default CD device if none is specified.
   if CdIo is NULL (we haven't initialized a specific device driver), 
@@ -1091,7 +1187,9 @@ cdio_open_win32 (const char *source_name)
     .free               = _cdio_win32_free,
     .get_arg            = _cdio_get_arg,
     .get_default_device = cdio_get_default_device_win32,
+    .get_devices        = cdio_get_devices_win32,
     .get_first_track_num= _cdio_get_first_track_num,
+    .get_mcn            = NULL, 
     .get_num_tracks     = _cdio_get_num_tracks,
     .get_track_format   = _cdio_get_track_format,
     .get_track_green    = _cdio_get_track_green,
@@ -1099,7 +1197,7 @@ cdio_open_win32 (const char *source_name)
     .get_track_msf      = _cdio_get_track_msf,
     .lseek              = NULL,
     .read               = NULL,
-    .read_audio_sector  = _cdio_read_raw_sector,
+    .read_audio_sectors = _cdio_read_audio_sectors,
     .read_mode2_sector  = _cdio_read_mode2_sector,
     .read_mode2_sectors = _cdio_read_mode2_sectors,
     .set_arg            = _cdio_set_arg,
