@@ -1,5 +1,5 @@
 /*
-    $Id: cdio.c,v 1.22 2003/08/31 14:26:06 rocky Exp $
+    $Id: cdio.c,v 1.36 2003/10/04 23:12:22 rocky Exp $
 
     Copyright (C) 2003 Rocky Bernstein <rocky@panix.com>
     Copyright (C) 2001 Herbert Valerio Riedel <hvr@gnu.org>
@@ -31,11 +31,13 @@
 #include <string.h>
 
 #include "cdio_assert.h"
+#include <cdio/cdio.h>
+#include <cdio/cd_types.h>
 #include <cdio/util.h>
 #include <cdio/logging.h>
 #include "cdio_private.h"
 
-static const char _rcsid[] = "$Id: cdio.c,v 1.22 2003/08/31 14:26:06 rocky Exp $";
+static const char _rcsid[] = "$Id: cdio.c,v 1.36 2003/10/04 23:12:22 rocky Exp $";
 
 
 const char *track_format2str[6] = 
@@ -74,6 +76,7 @@ CdIo_driver_t CdIo_all_drivers[CDIO_MAX_DRIVER+1] = {
    &cdio_have_false,
    NULL,
    NULL,
+   NULL,
    NULL
   },
 
@@ -84,7 +87,8 @@ CdIo_driver_t CdIo_all_drivers[CDIO_MAX_DRIVER+1] = {
    &cdio_have_bsdi,
    &cdio_open_bsdi,
    &cdio_get_default_device_bsdi,
-   &cdio_is_device_generic
+   &cdio_is_device_generic,
+   &cdio_get_devices_bsdi
   },
 
   {DRIVER_FREEBSD, 
@@ -94,17 +98,19 @@ CdIo_driver_t CdIo_all_drivers[CDIO_MAX_DRIVER+1] = {
    &cdio_have_freebsd,
    &cdio_open_freebsd,
    &cdio_get_default_device_freebsd,
-   &cdio_is_device_generic
+   &cdio_is_device_generic,
+   NULL
   },
 
   {DRIVER_LINUX, 
    CDIO_SRC_IS_DEVICE_MASK|CDIO_SRC_IS_NATIVE_MASK,
-   "Linux", 
-   "Linux ioctl and packet driver",
+   "GNU/Linux", 
+   "GNU/Linux ioctl and MMC driver",
    &cdio_have_linux,
    &cdio_open_linux,
    &cdio_get_default_device_linux,
-   &cdio_is_device_generic
+   &cdio_is_device_generic,
+   &cdio_get_devices_linux
   },
 
   {DRIVER_SOLARIS, 
@@ -114,7 +120,19 @@ CdIo_driver_t CdIo_all_drivers[CDIO_MAX_DRIVER+1] = {
    &cdio_have_solaris,
    &cdio_open_solaris,
    &cdio_get_default_device_solaris,
-   &cdio_is_device_generic
+   &cdio_is_device_generic,
+   &cdio_get_devices_solaris
+  },
+
+  {DRIVER_OSX, 
+   CDIO_SRC_IS_DEVICE_MASK|CDIO_SRC_IS_NATIVE_MASK|CDIO_SRC_IS_SCSI_MASK,
+   "OS X",
+   "Apple Darwin OS X driver",
+   &cdio_have_osx,
+   &cdio_open_osx,
+   &cdio_get_default_device_osx,
+   &cdio_is_device_generic,
+   &cdio_get_devices_osx
   },
 
   {DRIVER_WIN32, 
@@ -124,7 +142,8 @@ CdIo_driver_t CdIo_all_drivers[CDIO_MAX_DRIVER+1] = {
    &cdio_have_win32,
    &cdio_open_win32,
    &cdio_get_default_device_win32,
-   &cdio_is_device_win32
+   &cdio_is_device_win32,
+   &cdio_get_devices_win32
   },
 
   {DRIVER_BINCUE,
@@ -134,7 +153,8 @@ CdIo_driver_t CdIo_all_drivers[CDIO_MAX_DRIVER+1] = {
    &cdio_have_bincue,
    &cdio_open_bincue,
    &cdio_get_default_device_bincue,
-   NULL
+   NULL,
+   &cdio_get_devices_bincue
   },
 
   {DRIVER_NRG,
@@ -144,11 +164,35 @@ CdIo_driver_t CdIo_all_drivers[CDIO_MAX_DRIVER+1] = {
    &cdio_have_nrg,
    &cdio_open_nrg,
    &cdio_get_default_device_nrg,
-   NULL
+   NULL,
+   &cdio_get_devices_nrg
   }
 
 };
 
+static CdIo *
+scan_for_driver(driver_id_t start, driver_id_t end, const char *source_name) 
+{
+  driver_id_t driver_id;
+  
+  for (driver_id=start; driver_id<=end; driver_id++) {
+    if ((*CdIo_all_drivers[driver_id].have_driver)()) {
+      CdIo *ret=(*CdIo_all_drivers[driver_id].driver_open)(source_name);
+      if (ret != NULL) {
+        ret->driver_id = driver_id;
+        return ret;
+      }
+    }
+  }
+  return NULL;
+}
+
+const char *
+cdio_driver_describe(driver_id_t driver_id)
+{
+  return CdIo_all_drivers[driver_id].describe;
+}
+
 /*!
   Eject media in CD drive if there is a routine to do so. 
   Return 0 if success and 1 for failure, and 2 if no routine.
@@ -161,7 +205,7 @@ cdio_eject_media (CdIo **obj)
   if ((obj == NULL) || (*obj == NULL)) return 1;
 
   if ((*obj)->op.eject_media) {
-    int ret = (*obj)->op.eject_media ((*obj)->user_data);
+    int ret = (*obj)->op.eject_media ((*obj)->env);
     if (0 == ret) {
       cdio_destroy(*obj);
       *obj = NULL;
@@ -174,6 +218,18 @@ cdio_eject_media (CdIo **obj)
   }
 }
 
+/*!
+  Free device list returned by cdio_get_devices or
+  cdio_get_devices_with_cap.
+*/
+void cdio_free_device_list (char * device_list[]) 
+{
+  if (NULL == device_list) return;
+  for ( ; *device_list != NULL ; device_list++ )
+    free(*device_list);
+}
+
+
 /*!
   Return the value associatied with key. NULL is returned if obj is NULL
   or "key" does not exist.
@@ -184,7 +240,7 @@ cdio_get_arg (const CdIo *obj, const char key[])
   if (obj == NULL) return NULL;
   
   if (obj->op.get_arg) {
-    return obj->op.get_arg (obj->user_data, key);
+    return obj->op.get_arg (obj->env, key);
   } else {
     return NULL;
   }
@@ -219,10 +275,96 @@ cdio_get_default_device (const CdIo *obj)
   }
 }
 
-const char *
-cdio_driver_describe(driver_id_t driver_id)
+/*!Return an array of device names. If you want a specific
+  devices, dor a driver give that device, if you want hardware
+  devices, give DRIVER_DEVICE and if you want all possible devices,
+  image drivers and hardware drivers give DRIVER_UNKNOWN.
+  
+  NULL is returned if we couldn't return a list of devices.
+*/
+char **
+cdio_get_devices (driver_id_t driver_id)
 {
-  return CdIo_all_drivers[driver_id].describe;
+  CdIo *cdio;
+
+  switch (driver_id) {
+    /* FIXME: spit out unknown to give image drivers as well.  */
+  case DRIVER_UNKNOWN:
+  case DRIVER_DEVICE:
+    cdio = scan_for_driver(DRIVER_UNKNOWN, CDIO_MAX_DRIVER, NULL);
+    break;
+  default:
+    return (*CdIo_all_drivers[driver_id].get_devices)();
+  }
+  
+  if (cdio == NULL) return NULL;
+  if (cdio->op.get_devices) {
+    return cdio->op.get_devices ();
+  } else {
+    return NULL;
+  }
+}
+
+/*!
+  Return an array of device names in search_devices that have at
+  least the capabilities listed by cap.  If search_devices is NULL,
+  then we'll search all possible CD drives.  
+  
+  If "any" is set false then every capability listed in the extended
+  portion of capabilities (i.e. not the basic filesystem) must be
+  satisified. If "any" is set true, then if any of the capabilities
+  matches, we call that a success.
+  
+  To find a CD-drive of any type, use the mask CDIO_FS_MATCH_ALL.
+  
+  NULL is returned if we couldn't get a default device.
+*/
+char **
+cdio_get_devices_with_cap (char* search_devices[], 
+                           cdio_fs_anal_t capabilities, bool any)
+{
+  char **drives=search_devices;
+  char **drives_ret=NULL;
+  int num_drives=0;
+
+  if (NULL == drives) drives=cdio_get_devices(DRIVER_DEVICE);
+  if (NULL == drives) return NULL;
+
+  if (capabilities == CDIO_FS_MATCH_ALL) {
+    /* Duplicate drives into drives_ret. */
+    for( ; *drives != NULL; drives++ ) {
+      cdio_add_device_list(&drives_ret, *drives, &num_drives);
+    }
+  } else {
+    cdio_fs_anal_t got_fs=0;
+    cdio_fs_anal_t need_fs = CDIO_FSTYPE(capabilities);
+    cdio_fs_anal_t need_fs_ext;
+    need_fs_ext = capabilities & ~CDIO_FS_MASK;
+      
+    for( ;  *drives != NULL; drives++ ) {
+      CdIo *cdio = cdio_open(*drives, DRIVER_UNKNOWN);
+      
+      if (NULL != cdio) {
+        track_t first_track = cdio_get_first_track_num(cdio);
+        cdio_analysis_t cdio_analysis; 
+        got_fs = cdio_guess_cd_type(cdio, 0, first_track, 
+                                &cdio_analysis);
+        /* Match on fs and add */
+        if ( (CDIO_FS_UNKNOWN == need_fs || CDIO_FSTYPE(got_fs) == need_fs) )
+          {
+            bool doit = any
+              ? (got_fs & need_fs_ext)  != 0
+              : (got_fs | ~need_fs_ext) == -1;
+            if (doit) 
+              cdio_add_device_list(&drives_ret, *drives, &num_drives);
+          }
+             
+        cdio_destroy(cdio);
+      }
+    }
+    cdio_add_device_list(&drives_ret, NULL, &num_drives);
+  }
+  return drives_ret;
 }
 
 /*!
@@ -231,9 +373,9 @@ cdio_driver_describe(driver_id_t driver_id)
   then return NULL.
 */
 const char *
-cdio_get_driver_name (const CdIo *obj) 
+cdio_get_driver_name (const CdIo *cdio) 
 {
-  return CdIo_all_drivers[obj->driver_id].name;
+  return CdIo_all_drivers[cdio->driver_id].name;
 }
 
 
@@ -242,28 +384,43 @@ cdio_get_driver_name (const CdIo *obj)
   CDIO_INVALID_TRACK is returned on error.
 */
 track_t
-cdio_get_first_track_num (const CdIo *obj)
+cdio_get_first_track_num (const CdIo *cdio)
 {
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
 
-  if (obj->op.get_first_track_num) {
-    return obj->op.get_first_track_num (obj->user_data);
+  if (cdio->op.get_first_track_num) {
+    return cdio->op.get_first_track_num (cdio->env);
   } else {
     return CDIO_INVALID_TRACK;
   }
 }
 
+/*!
+  Return a string containing the name of the driver in use.
+  if CdIo is NULL (we haven't initialized a specific device driver), 
+  then return NULL.
+*/
+char *
+cdio_get_mcn (const CdIo *cdio) 
+{
+  if (cdio->op.get_mcn) {
+    return cdio->op.get_mcn (cdio->env);
+  } else {
+    return NULL;
+  }
+}
+
 /*! 
   Return the number of tracks in the current medium.
   CDIO_INVALID_TRACK is returned on error.
 */
 track_t
-cdio_get_num_tracks (const CdIo *obj)
+cdio_get_num_tracks (const CdIo *cdio)
 {
-  if (obj == NULL) return CDIO_INVALID_TRACK;
+  if (cdio == NULL) return CDIO_INVALID_TRACK;
 
-  if (obj->op.get_num_tracks) {
-    return obj->op.get_num_tracks (obj->user_data);
+  if (cdio->op.get_num_tracks) {
+    return cdio->op.get_num_tracks (cdio->env);
   } else {
     return CDIO_INVALID_TRACK;
   }
@@ -273,12 +430,12 @@ cdio_get_num_tracks (const CdIo *obj)
   Get format of track. 
 */
 track_format_t
-cdio_get_track_format(const CdIo *obj, track_t track_num)
+cdio_get_track_format(const CdIo *cdio, track_t track_num)
 {
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
 
-  if (obj->op.get_track_format) {
-    return obj->op.get_track_format (obj->user_data, track_num);
+  if (cdio->op.get_track_format) {
+    return cdio->op.get_track_format (cdio->env, track_num);
   } else {
     return TRACK_FORMAT_ERROR;
   }
@@ -293,12 +450,12 @@ cdio_get_track_format(const CdIo *obj, track_t track_num)
   FIXME: there's gotta be a better design for this and get_track_format?
 */
 bool
-cdio_get_track_green(const CdIo *obj, track_t track_num)
+cdio_get_track_green(const CdIo *cdio, track_t track_num)
 {
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
 
-  if (obj->op.get_track_green) {
-    return obj->op.get_track_green (obj->user_data, track_num);
+  if (cdio->op.get_track_green) {
+    return cdio->op.get_track_green (cdio->env, track_num);
   } else {
     return false;
   }
@@ -306,43 +463,44 @@ cdio_get_track_green(const CdIo *obj, track_t track_num)
 
 /*!  
   Return the starting LBA for track number
-  track_num in obj.  Tracks numbers start at 1.
+  track_num in cdio.  Tracks numbers start at 1.
   The "leadout" track is specified either by
   using track_num LEADOUT_TRACK or the total tracks+1.
   CDIO_INVALID_LBA is returned on error.
 */
 lba_t
-cdio_get_track_lba(const CdIo *obj, track_t track_num)
+cdio_get_track_lba(const CdIo *cdio, track_t track_num)
 {
-  if (obj == NULL) return CDIO_INVALID_LBA;
+  if (cdio == NULL) return CDIO_INVALID_LBA;
 
-  if (obj->op.get_track_lba) {
-    return obj->op.get_track_lba (obj->user_data, track_num);
+  if (cdio->op.get_track_lba) {
+    return cdio->op.get_track_lba (cdio->env, track_num);
   } else {
     msf_t msf;
-    if (cdio_get_track_msf(obj, track_num, &msf))
-      return cdio_msf_to_lba(&msf);
+    if (cdio->op.get_track_msf) 
+      if (cdio_get_track_msf(cdio, track_num, &msf))
+        return cdio_msf_to_lba(&msf);
     return CDIO_INVALID_LBA;
   }
 }
 
 /*!  
   Return the starting LSN for track number
-  track_num in obj.  Tracks numbers start at 1.
+  track_num in cdio.  Tracks numbers start at 1.
   The "leadout" track is specified either by
   using track_num LEADOUT_TRACK or the total tracks+1.
   CDIO_INVALID_LBA is returned on error.
 */
 lsn_t
-cdio_get_track_lsn(const CdIo *obj, track_t track_num)
+cdio_get_track_lsn(const CdIo *cdio, track_t track_num)
 {
-  if (obj == NULL) return CDIO_INVALID_LBA;
+  if (cdio == NULL) return CDIO_INVALID_LBA;
 
-  if (obj->op.get_track_lba) {
-    return cdio_lba_to_lsn(obj->op.get_track_lba (obj->user_data, track_num));
+  if (cdio->op.get_track_lba) {
+    return cdio_lba_to_lsn(cdio->op.get_track_lba (cdio->env, track_num));
   } else {
     msf_t msf;
-    if (cdio_get_track_msf(obj, track_num, &msf))
+    if (cdio_get_track_msf(cdio, track_num, &msf))
       return cdio_msf_to_lsn(&msf);
     return CDIO_INVALID_LSN;
   }
@@ -350,18 +508,23 @@ cdio_get_track_lsn(const CdIo *obj, track_t track_num)
 
 /*!  
   Return the starting MSF (minutes/secs/frames) for track number
-  track_num in obj.  Track numbers start at 1.
+  track_num in cdio.  Track numbers start at 1.
   The "leadout" track is specified either by
   using track_num LEADOUT_TRACK or the total tracks+1.
   False is returned if there is no track entry.
 */
 bool
-cdio_get_track_msf(const CdIo *obj, track_t track_num, /*out*/ msf_t *msf)
+cdio_get_track_msf(const CdIo *cdio, track_t track_num, /*out*/ msf_t *msf)
 {
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
 
-  if (obj->op.get_track_msf) {
-    return obj->op.get_track_msf (obj->user_data, track_num, msf);
+  if (cdio->op.get_track_msf) {
+    return cdio->op.get_track_msf (cdio->env, track_num, msf);
+  } else if (cdio->op.get_track_lba) {
+    lba_t lba = cdio->op.get_track_lba (cdio->env, track_num);
+    if (lba  == CDIO_INVALID_LBA) return false;
+    cdio_lba_to_msf(lba, msf);
+    return true;
   } else {
     return false;
   }
@@ -374,13 +537,13 @@ cdio_get_track_msf(const CdIo *obj, track_t track_num, /*out*/ msf_t *msf)
   0 is returned if there is an error.
 */
 unsigned int
-cdio_get_track_sec_count(const CdIo *obj, track_t track_num)
+cdio_get_track_sec_count(const CdIo *cdio, track_t track_num)
 {
-  track_t num_tracks = cdio_get_num_tracks(obj);
+  track_t num_tracks = cdio_get_num_tracks(cdio);
 
   if (track_num >=1 && track_num <= num_tracks) 
-    return ( cdio_get_track_lba(obj, track_num+1) 
-             - cdio_get_track_lba(obj, track_num) );
+    return ( cdio_get_track_lba(cdio, track_num+1) 
+             - cdio_get_track_lba(cdio, track_num) );
   return 0;
 }
 
@@ -427,30 +590,30 @@ cdio_init(void)
 }
 
 CdIo *
-cdio_new (void *user_data, const cdio_funcs *funcs)
+cdio_new (void *env, const cdio_funcs *funcs)
 {
-  CdIo *new_obj;
+  CdIo *new_cdio;
 
-  new_obj = _cdio_malloc (sizeof (CdIo));
+  new_cdio = _cdio_malloc (sizeof (CdIo));
 
-  new_obj->user_data = user_data;
-  new_obj->op = *funcs;
+  new_cdio->env = env;
+  new_cdio->op = *funcs;
 
-  return new_obj;
+  return new_cdio;
 }
 
 /*!
-  Free any resources associated with obj.
+  Free any resources associated with cdio.
 */
 void
-cdio_destroy (CdIo *obj)
+cdio_destroy (CdIo *cdio)
 {
   CdIo_last_driver = CDIO_DRIVER_UNINIT;
-  if (obj == NULL) return;
+  if (cdio == NULL) return;
 
-  if (obj->op.free != NULL) 
-    obj->op.free (obj->user_data);
-  free (obj);
+  if (cdio->op.free != NULL) 
+    cdio->op.free (cdio->env);
+  free (cdio);
 }
 
 /*!
@@ -459,12 +622,12 @@ cdio_destroy (CdIo *obj)
   Similar to (if not the same as) libc's lseek()
 */
 off_t
-cdio_lseek (CdIo *obj, off_t offset, int whence)
+cdio_lseek (const CdIo *cdio, off_t offset, int whence)
 {
-  if (obj == NULL) return -1;
+  if (cdio == NULL) return -1;
   
-  if (obj->op.lseek)
-    return obj->op.lseek (obj->user_data, offset, whence);
+  if (cdio->op.lseek)
+    return cdio->op.lseek (cdio->env, offset, whence);
   return -1;
 }
 
@@ -474,23 +637,35 @@ cdio_lseek (CdIo *obj, off_t offset, int whence)
   Similar to (if not the same as) libc's read()
 */
 ssize_t
-cdio_read (CdIo *obj, void *buf, size_t size)
+cdio_read (const CdIo *cdio, void *buf, size_t size)
 {
-  if (obj == NULL) return -1;
+  if (cdio == NULL) return -1;
   
-  if (obj->op.read)
-    return obj->op.read (obj->user_data, buf, size);
+  if (cdio->op.read)
+    return cdio->op.read (cdio->env, buf, size);
   return -1;
 }
 
 int
-cdio_read_audio_sector (CdIo *obj, void *buf, lsn_t lsn) 
+cdio_read_audio_sector (const CdIo *cdio, void *buf, lsn_t lsn) 
 {
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
   cdio_assert (buf != NULL);
 
-  if  (obj->op.read_audio_sector != NULL)
-    return obj->op.read_audio_sector (obj->user_data, buf, lsn);
+  if  (cdio->op.read_audio_sectors != NULL)
+    return cdio->op.read_audio_sectors (cdio->env, buf, lsn, 1);
+  return -1;
+}
+
+int
+cdio_read_audio_sectors (const CdIo *cdio, void *buf, lsn_t lsn,
+                         unsigned int nblocks) 
+{
+  cdio_assert (cdio != NULL);
+  cdio_assert (buf != NULL);
+
+  if  (cdio->op.read_audio_sectors != NULL)
+    return cdio->op.read_audio_sectors (cdio->env, buf, lsn, nblocks);
   return -1;
 }
 
@@ -499,24 +674,24 @@ cdio_read_audio_sector (CdIo *obj, void *buf, lsn_t lsn)
    into data starting from lsn. Returns 0 if no error. 
  */
 int
-cdio_read_mode1_sector (CdIo *obj, void *data, lsn_t lsn, bool is_form2)
+cdio_read_mode1_sector (const CdIo *cdio, void *data, lsn_t lsn, bool is_form2)
 {
   uint32_t size = is_form2 ? M2RAW_SECTOR_SIZE : CDIO_CD_FRAMESIZE ;
   char buf[M2RAW_SECTOR_SIZE] = { 0, };
   int ret;
   
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
   cdio_assert (data != NULL);
 
-  if (obj->op.lseek && obj->op.read) {
-    if (0 > cdio_lseek(obj, CDIO_CD_FRAMESIZE*lsn, SEEK_SET))
+  if (cdio->op.lseek && cdio->op.read) {
+    if (0 > cdio_lseek(cdio, CDIO_CD_FRAMESIZE*lsn, SEEK_SET))
       return -1;
-    if (0 > cdio_read(obj, buf, CDIO_CD_FRAMESIZE))
+    if (0 > cdio_read(cdio, buf, CDIO_CD_FRAMESIZE))
       return -1;
     memcpy (data, buf, size);
     return 0;
   } else {
-    ret = cdio_read_mode2_sector(obj, data, lsn, is_form2);
+    ret = cdio_read_mode2_sector(cdio, data, lsn, is_form2);
     if (ret == 0) 
       memcpy (data, buf+CDIO_CD_SUBHEADER_SIZE, size);
   }
@@ -525,17 +700,17 @@ cdio_read_mode1_sector (CdIo *obj, void *data, lsn_t lsn, bool is_form2)
 }
 
 int
-cdio_read_mode1_sectors (CdIo *obj, void *data, lsn_t lsn, bool is_form2, 
-                         unsigned int num_sectors)
+cdio_read_mode1_sectors (const CdIo *cdio, void *data, lsn_t lsn, 
+                         bool is_form2,  unsigned int num_sectors)
 {
   uint32_t size = is_form2 ? M2RAW_SECTOR_SIZE : CDIO_CD_FRAMESIZE ;
   int retval;
   int i;
 
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
   
   for (i = 0; i < num_sectors; i++) {
-    if ( (retval = cdio_read_mode1_sector (obj, 
+    if ( (retval = cdio_read_mode1_sector (cdio, 
                                            ((char *)data) + (size * i),
                                            lsn + i, is_form2)) )
       return retval;
@@ -548,87 +723,63 @@ cdio_read_mode1_sectors (CdIo *obj, void *data, lsn_t lsn, bool is_form2,
    from lsn. Returns 0 if no error. 
  */
 int
-cdio_read_mode2_sector (CdIo *obj, void *buf, uint32_t lsn, bool is_form2)
+cdio_read_mode2_sector (const CdIo *cdio, void *buf, lsn_t lsn, 
+                        bool is_form2)
 {
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
   cdio_assert (buf != NULL);
-  cdio_assert (obj->op.read_mode2_sector != NULL 
-	      || obj->op.read_mode2_sectors != NULL);
+  cdio_assert (cdio->op.read_mode2_sector != NULL 
+	      || cdio->op.read_mode2_sectors != NULL);
 
-  if (obj->op.read_mode2_sector)
-    return obj->op.read_mode2_sector (obj->user_data, buf, lsn, is_form2);
+  if (cdio->op.read_mode2_sector)
+    return cdio->op.read_mode2_sector (cdio->env, buf, lsn, is_form2);
 
   /* fallback */
-  if (obj->op.read_mode2_sectors != NULL)
-    return cdio_read_mode2_sectors (obj, buf, lsn, is_form2, 1);
+  if (cdio->op.read_mode2_sectors != NULL)
+    return cdio_read_mode2_sectors (cdio, buf, lsn, is_form2, 1);
   return 1;
 }
 
 int
-cdio_read_mode2_sectors (CdIo *obj, void *buf, lsn_t lsn, bool mode2raw, 
+cdio_read_mode2_sectors (const CdIo *cdio, void *buf, lsn_t lsn, bool mode2raw, 
                          unsigned num_sectors)
 {
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
   cdio_assert (buf != NULL);
-  cdio_assert (obj->op.read_mode2_sectors != NULL);
+  cdio_assert (cdio->op.read_mode2_sectors != NULL);
   
-  return obj->op.read_mode2_sectors (obj->user_data, buf, lsn,
+  return cdio->op.read_mode2_sectors (cdio->env, buf, lsn,
                                      mode2raw, num_sectors);
 }
 
 uint32_t
-cdio_stat_size (CdIo *obj)
+cdio_stat_size (const CdIo *cdio)
 {
-  cdio_assert (obj != NULL);
+  cdio_assert (cdio != NULL);
 
-  return obj->op.stat_size (obj->user_data);
+  return cdio->op.stat_size (cdio->env);
 }
 
 /*!
   Set the arg "key" with "value" in the source device.
 */
 int
-cdio_set_arg (CdIo *obj, const char key[], const char value[])
+cdio_set_arg (CdIo *cdio, const char key[], const char value[])
 {
-  cdio_assert (obj != NULL);
-  cdio_assert (obj->op.set_arg != NULL);
+  cdio_assert (cdio != NULL);
+  cdio_assert (cdio->op.set_arg != NULL);
   cdio_assert (key != NULL);
 
-  return obj->op.set_arg (obj->user_data, key, value);
-}
-
-static CdIo *
-scan_for_driver(driver_id_t start, driver_id_t end, const char *source_name) 
-{
-  driver_id_t driver_id;
-  
-  for (driver_id=start; driver_id<=end; driver_id++) {
-    if ((*CdIo_all_drivers[driver_id].have_driver)()) {
-      CdIo *ret=(*CdIo_all_drivers[driver_id].driver_open)(source_name);
-      if (ret != NULL) {
-        ret->driver_id = driver_id;
-        return ret;
-      }
-    }
-  }
-  return NULL;
+  return cdio->op.set_arg (cdio->env, key, value);
 }
 
 /*! Sets up to read from place specified by source_name and
-  driver_id This should be called before using any other routine,
+  driver_id. This should be called before using any other routine,
   except cdio_init. This will call cdio_init, if that hasn't been
-  done previously.  to call one of the specific routines below. 
+  done previously.
   
   NULL is returned on error.
 */
-/* In the future we'll have more complicated code to allow selection
-   of an I/O routine as well as code to find an appropriate default
-   routine among the "registered" routines. Possibly classes too
-   disk-based, SCSI-based, native-based, vendor (e.g. Sony, or
-   Plextor) based 
-
-   For now though, we'll start more simply...
-*/
 CdIo *
 cdio_open (const char *orig_source_name, driver_id_t driver_id)
 {
@@ -636,7 +787,7 @@ cdio_open (const char *orig_source_name, driver_id_t driver_id)
   
   if (CdIo_last_driver == -1) cdio_init();
 
-  if (NULL == orig_source_name) 
+  if (NULL == orig_source_name || strlen(orig_source_name)==0) 
     source_name = cdio_get_default_device(NULL);
   else 
     source_name = strdup(orig_source_name);
@@ -692,12 +843,12 @@ cdio_open (const char *orig_source_name, driver_id_t driver_id)
   case DRIVER_LINUX:
   case DRIVER_SOLARIS:
   case DRIVER_WIN32:
+  case DRIVER_OSX:
   case DRIVER_NRG:
   case DRIVER_BINCUE:
     if ((*CdIo_all_drivers[driver_id].have_driver)()) {
       CdIo *ret = (*CdIo_all_drivers[driver_id].driver_open)(source_name);
-      ret->driver_id = driver_id;
-      free(source_name);
+      if (ret) ret->driver_id = driver_id;
       return ret;
     }
   }
