@@ -1,5 +1,5 @@
 /*
-    $Id: iso9660.c,v 1.7 2003/09/01 15:10:43 rocky Exp $
+    $Id: iso9660.c,v 1.15 2003/09/21 07:43:39 rocky Exp $
 
     Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>
     Copyright (C) 2003 Rocky Bernstein <rocky@panix.com>
@@ -19,9 +19,14 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-#ifdef HAVE_CONFIG_H
-# include "config.h"
-#endif
+/* Private headers */
+#include "iso9660_private.h"
+#include "cdio_assert.h"
+#include "bytesex.h"
+
+/* Public headers */
+#include <cdio/iso9660.h>
+#include <cdio/util.h>
 
 #include <time.h>
 #include <ctype.h>
@@ -32,16 +37,7 @@
 #include <stdio.h>
 #endif
 
-/* Public headers */
-#include <cdio/iso9660.h>
-#include <cdio/util.h>
-
-/* Private headers */
-#include "iso9660_private.h"
-#include "cdio_assert.h"
-#include "bytesex.h"
-
-static const char _rcsid[] = "$Id: iso9660.c,v 1.7 2003/09/01 15:10:43 rocky Exp $";
+static const char _rcsid[] = "$Id: iso9660.c,v 1.15 2003/09/21 07:43:39 rocky Exp $";
 
 /* some parameters... */
 #define SYSTEM_ID         "CD-RTOS CD-BRIDGE"
@@ -51,31 +47,92 @@ static void
 pathtable_get_size_and_entries(const void *pt, unsigned int *size, 
                                unsigned int *entries);
 
-static void
-_idr_set_time (uint8_t _idr_date[], const struct tm *_tm)
+/*!
+  Get time structure from structure in an ISO 9660 directory index 
+  record. Even though tm_wday and tm_yday fields are not explicitly in
+  idr_date, the are calculated from the other fields.
+
+  If tm is to reflect the localtime set use_localtime true, otherwise
+  tm will reported in GMT.
+*/
+void
+iso9660_get_dtime (const iso9660_dtime_t *idr_date, bool use_localtime,
+                   /*out*/ struct tm *tm)
 {
-  memset (_idr_date, 0, 7);
+  time_t t;
+  struct tm *temp_tm;
+  
+  if (!idr_date) return;
 
-  if (!_tm)
-    return;
+  tm->tm_year   = idr_date->dt_year;
+  tm->tm_mon    = idr_date->dt_month - 1;
+  tm->tm_mday   = idr_date->dt_day;
+  tm->tm_hour   = idr_date->dt_hour;
+  tm->tm_min    = idr_date->dt_minute;
+  tm->tm_sec    = idr_date->dt_second;
 
-  _idr_date[0] = _tm->tm_year;
-  _idr_date[1] = _tm->tm_mon + 1;
-  _idr_date[2] = _tm->tm_mday;
-  _idr_date[3] = _tm->tm_hour;
-  _idr_date[4] = _tm->tm_min;
-  _idr_date[5] = _tm->tm_sec;
-  _idr_date[6] = 0x00; /* tz, GMT -48 +52 in 15min intervals */
+  /* Recompute tm_wday and tm_yday via mktime. */
+  t = mktime(tm);
+
+  if (use_localtime)
+    temp_tm = localtime(&t);
+  else
+    temp_tm = gmtime(&t);
+
+  memcpy(tm, temp_tm, sizeof(struct tm));
 }
 
-static void
-_pvd_set_time (char _pvd_date[], const struct tm *_tm)
+/*!
+  Set time in format used in ISO 9660 directory index record
+  from a Unix time structure. */
+void
+iso9660_set_dtime (const struct tm *tm, /*out*/ iso9660_dtime_t *idr_date)
 {
+  memset (idr_date, 0, 7);
+
+  if (!tm) return;
+
+  idr_date->dt_year   = tm->tm_year;
+  idr_date->dt_month  = tm->tm_mon + 1;
+  idr_date->dt_day    = tm->tm_mday;
+  idr_date->dt_hour   = tm->tm_hour;
+  idr_date->dt_minute = tm->tm_min;
+  idr_date->dt_second = tm->tm_sec;
+
+#ifdef HAVE_TM_GMTOFF
+  /* The ISO 9660 timezone is in the range -48..+52 and each unit
+     represents a 15-minute interval. */
+  idr_date->dt_gmtoff = tm->tm_gmtoff / (15 * 60);
+
+  if (tm->tm_isdst) idr_date->dt_gmtoff -= 4;
+
+  if (idr_date->dt_gmtoff < -48 ) {
+    
+    cdio_warn ("Converted ISO 9660 timezone %d is less than -48. Adjusted", 
+               idr_date->dt_gmtoff);
+    idr_date->dt_gmtoff = -48;
+  } else if (idr_date->dt_gmtoff > 52) {
+    cdio_warn ("Converted ISO 9660 timezone %d is over 52. Adjusted", 
+               idr_date->dt_gmtoff);
+    idr_date->dt_gmtoff = 52;
+  }
+#else 
+  idr_date->dt_gmtoff = 0;
+#endif
+}
+
+/*!
+  Set "long" time in format used in ISO 9660 primary volume descriptor
+  from a Unix time structure. */
+void
+iso9660_set_ltime (const struct tm *_tm, /*out*/ iso9660_ltime_t *pvd_date)
+{
+  char *_pvd_date = (char *) pvd_date; 
+
   memset (_pvd_date, '0', 16);
   _pvd_date[16] = (int8_t) 0; /* tz */
 
-  if (!_tm)
-    return;
+  if (!_tm) return;
 
   snprintf(_pvd_date, 17, 
            "%4.4d%2.2d%2.2d" "%2.2d%2.2d%2.2d" "%2.2d",
@@ -86,6 +143,46 @@ _pvd_set_time (char _pvd_date[], const struct tm *_tm)
   _pvd_date[16] = (int8_t) 0; /* tz */
 }
 
+/*!
+   Convert ISO-9660 file name that stored in a directory entry into 
+   what's usually listed as the file name in a listing.
+   Lowercase name, and trailing ;1's or .;1's and turn the
+   other ;'s into version numbers.
+
+   The length of the translated string is returned.
+*/
+int 
+iso9660_name_translate(const char *old, char *new)
+{
+  int len = strlen(old);
+  int i;
+  
+  for (i = 0; i < len; i++) {
+    unsigned char c = old[i];
+    if (!c)
+      break;
+    
+    /* lower case */
+    if (isupper(c)) c = tolower(c);	
+    
+    /* Drop trailing '.;1' (ISO 9660:1988 7.5.1 requires period) */
+    if (c == '.' && i == len - 3 && old[i + 1] == ';' && old[i + 2] == '1')
+      break;
+    
+    /* Drop trailing ';1' */
+    if (c == ';' && i == len - 2 && old[i + 1] == '1')
+      break;
+    
+    /* Convert remaining ';' to '.' */
+    if (c == ';')
+      c = '.';
+    
+    new[i] = c;
+  }
+  new[i] = '\0';
+  return i;
+}
+
 /*!  
   Pad string src with spaces to size len and copy this to dst. If
   len is less than the length of src, dst will be truncated to the
@@ -212,21 +309,19 @@ iso9660_set_evd(void *pd)
   memcpy(pd, &ied, sizeof(ied));
 }
 
-/* important date to celebrate (for me at least =)
-   -- until user customization is implemented... */
-static const time_t _vcd_time = 269222400L;
-                                       
 void
 iso9660_set_pvd(void *pd,
-            const char volume_id[],
-            const char publisher_id[],
-            const char preparer_id[],
-            const char application_id[],
-            uint32_t iso_size,
-            const void *root_dir,
-            uint32_t path_table_l_extent,
-            uint32_t path_table_m_extent,
-            uint32_t path_table_size)
+                const char volume_id[],
+                const char publisher_id[],
+                const char preparer_id[],
+                const char application_id[],
+                uint32_t iso_size,
+                const void *root_dir,
+                uint32_t path_table_l_extent,
+                uint32_t path_table_m_extent,
+                uint32_t path_table_size,
+                const time_t *pvd_time
+                )
 {
   iso9660_pvd_t ipd;
 
@@ -272,10 +367,10 @@ iso9660_set_pvd(void *pd,
   iso9660_strncpy_pad (ipd.abstract_file_id     , "", 37, ISO9660_DCHARS);
   iso9660_strncpy_pad (ipd.bibliographic_file_id, "", 37, ISO9660_DCHARS);
 
-  _pvd_set_time (ipd.creation_date, gmtime (&_vcd_time));
-  _pvd_set_time (ipd.modification_date, gmtime (&_vcd_time));
-  _pvd_set_time (ipd.expiration_date, NULL);
-  _pvd_set_time (ipd.effective_date, NULL);
+  iso9660_set_ltime (gmtime (pvd_time), &(ipd.creation_date));
+  iso9660_set_ltime (gmtime (pvd_time), &(ipd.modification_date));
+  iso9660_set_ltime (NULL,              &(ipd.expiration_date));
+  iso9660_set_ltime (NULL,              &(ipd.effective_date));
 
   ipd.file_structure_version = to_711(1);
 
@@ -302,19 +397,19 @@ iso9660_dir_calc_record_size(unsigned int namelen, unsigned int su_len)
 
 void
 iso9660_dir_add_entry_su(void *dir,
-                         const char name[], 
+                         const char filename[], 
                          uint32_t extent,
                          uint32_t size,
-                         uint8_t flags,
+                         uint8_t file_flags,
                          const void *su_data,
-                         unsigned int su_size)
+                         unsigned int su_size,
+                         const time_t *entry_time)
 {
   iso9660_dir_t *idr = dir;
   uint8_t *dir8 = dir;
   unsigned int offset = 0;
   uint32_t dsize = from_733(idr->size);
   int length, su_offset;
-
   cdio_assert (sizeof(iso9660_dir_t) == 33);
 
   if (!dsize && !idr->length)
@@ -323,11 +418,11 @@ iso9660_dir_add_entry_su(void *dir,
   cdio_assert (dsize > 0 && !(dsize % ISO_BLOCKSIZE));
   cdio_assert (dir != NULL);
   cdio_assert (extent > 17);
-  cdio_assert (name != NULL);
-  cdio_assert (strlen(name) <= MAX_ISOPATHNAME);
+  cdio_assert (filename != NULL);
+  cdio_assert (strlen(filename) <= MAX_ISOPATHNAME);
 
   length = sizeof(iso9660_dir_t);
-  length += strlen(name);
+  length += strlen(filename);
   length = _cdio_ceil2block (length, 2); /* pad to word boundary */
   su_offset = length;
   length += su_size;
@@ -371,15 +466,16 @@ iso9660_dir_add_entry_su(void *dir,
   idr->extent = to_733(extent);
   idr->size = to_733(size);
   
-  _idr_set_time (idr->date, gmtime (&_vcd_time));
+  iso9660_set_dtime (gmtime(entry_time), &(idr->recording_time));
   
-  idr->flags = to_711(flags);
+  idr->file_flags = to_711(file_flags);
 
   idr->volume_sequence_number = to_723(1);
 
-  idr->name_len = to_711(strlen(name) ? strlen(name) : 1); /* working hack! */
+  idr->filename_len = to_711(strlen(filename) 
+                             ? strlen(filename) : 1); /* working hack! */
 
-  memcpy(idr->name, name, from_711(idr->name_len));
+  memcpy(idr->filename, filename, from_711(idr->filename_len));
   memcpy(&dir8[offset] + su_offset, su_data, su_size);
 }
 
@@ -388,9 +484,11 @@ iso9660_dir_init_new (void *dir,
                       uint32_t self,
                       uint32_t ssize,
                       uint32_t parent,
-                      uint32_t psize)
+                      uint32_t psize,
+                      const time_t *dir_time)
 {
-  iso9660_dir_init_new_su (dir, self, ssize, NULL, 0, parent, psize, NULL, 0);
+  iso9660_dir_init_new_su (dir, self, ssize, NULL, 0, parent, psize, NULL, 
+                           0, dir_time);
 }
 
 void 
@@ -402,7 +500,8 @@ iso9660_dir_init_new_su (void *dir,
                          uint32_t parent,
                          uint32_t psize,
                          const void *psu_data,
-                         unsigned int psu_size)
+                         unsigned int psu_size,
+                         const time_t *dir_time)
 {
   cdio_assert (ssize > 0 && !(ssize % ISO_BLOCKSIZE));
   cdio_assert (psize > 0 && !(psize % ISO_BLOCKSIZE));
@@ -412,12 +511,13 @@ iso9660_dir_init_new_su (void *dir,
 
   /* "\0" -- working hack due to padding  */
   iso9660_dir_add_entry_su (dir, "\0", self, ssize, ISO_DIRECTORY, ssu_data, 
-                            ssu_size); 
+                            ssu_size, dir_time); 
 
   iso9660_dir_add_entry_su (dir, "\1", parent, psize, ISO_DIRECTORY, psu_data, 
-                            psu_size);
+                            psu_size, dir_time);
 }
 
+/* Zero's out pathable. Do this first.  */
 void 
 iso9660_pathtable_init (void *pt)
 {
@@ -750,6 +850,9 @@ iso9660_get_pvd_block_size(const iso9660_pvd_t *pvd)
   return from_723(pvd->logical_block_size);
 }
 
+/*! Return the primary volume id version number (of pvd).
+    If there is an error 0 is returned. 
+ */
 int
 iso9660_get_pvd_version(const iso9660_pvd_t *pvd) 
 {
@@ -757,6 +860,9 @@ iso9660_get_pvd_version(const iso9660_pvd_t *pvd)
   return pvd->version;
 }
 
+/*! Return the LSN of the root directory for pvd.
+    If there is an error CDIO_INVALID_LSN is returned. 
+ */
 lsn_t
 iso9660_get_root_lsn(const iso9660_pvd_t *pvd) 
 {
