@@ -1,5 +1,5 @@
 /*
-    $Id: cd-info.c,v 1.31 2003/09/01 22:31:46 rocky Exp $
+    $Id: cd-info.c,v 1.43 2003/10/15 01:59:51 rocky Exp $
 
     Copyright (C) 2003 Rocky Bernstein <rocky@panix.com>
     Copyright (C) 1996,1997,1998  Gerd Knorr <kraxel@bytesex.org>
@@ -24,20 +24,8 @@
   the CD.
  
 */
-#define PROGRAM_NAME "CD Info"
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <unistd.h>
-#include <string.h>
-
-#include <popt.h>
-/* Accomodate to older popt that doesn't support the "optional" flag */
-#ifndef POPT_ARGFLAG_OPTIONAL
-#define POPT_ARGFLAG_OPTIONAL 0
-#endif
-
-#include "config.h"
+#include "util.h"
 
 #ifdef HAVE_CDDB
 #include <cddb/cddb.h>
@@ -49,8 +37,6 @@
 #include <libvcd/info.h>
 #endif
 
-#include <cdio/cdio.h>
-#include <cdio/logging.h>
 #include <cdio/util.h>
 #include <cdio/cd_types.h>
 #include <cdio/iso9660.h>
@@ -71,41 +57,6 @@
  
 #include <errno.h>
 
-#ifdef ENABLE_NLS
-#include <locale.h>
-#    include <libintl.h>
-#    define _(String) dgettext ("cdinfo", String)
-#else
-/* Stubs that do something close enough.  */
-#    define _(String) (String)
-#endif
-
-/* The following test is to work around the gross typo in
-   systems like Sony NEWS-OS Release 4.0C, whereby EXIT_FAILURE
-   is defined to 0, not 1.  */
-#if !EXIT_FAILURE
-# undef EXIT_FAILURE
-# define EXIT_FAILURE 1
-#endif
-
-#ifndef EXIT_SUCCESS
-# define EXIT_SUCCESS 0
-#endif
-
-#define DEBUG 1
-#if DEBUG
-#define dbg_print(level, s, args...) \
-   if (opts.debug_level >= level) \
-     fprintf(stderr, "%s: "s, __func__ , ##args)
-#else
-#define dbg_print(level, s, args...) 
-#endif
-
-#define err_exit(fmt, args...) \
-  fprintf(stderr, "%s: "fmt, program_name, ##args); \
-  myexit(cdio, EXIT_FAILURE)		     
-  
-
 #if 0
 #define STRONG "\033[1m"
 #define NORMAL "\033[0m"
@@ -115,27 +66,10 @@
 #endif
 
 #if CDIO_IOCTL_FINISHED
-struct cdrom_mcn           mcn;
 struct cdrom_multisession  ms;
 struct cdrom_subchnl       sub;
 #endif
 
-char *source_name = NULL;
-char *program_name;
-
-const char *argp_program_version     = PROGRAM_NAME " " VERSION;
-const char *argp_program_bug_address = "rocky@panix.com";
-
-typedef enum
-{
-  IMAGE_AUTO,
-  IMAGE_DEVICE,
-  IMAGE_BIN,
-  IMAGE_CUE,
-  IMAGE_NRG,
-  IMAGE_UNKNOWN
-} source_image_t;
-
 /* Used by `main' to communicate with `parse_opt'. And global options
  */
 struct arguments
@@ -143,6 +77,7 @@ struct arguments
   int            no_tracks;
   int            no_ioctl;
   int            no_analysis;
+  char          *access_mode; /* Access method driver should use for control */
 #ifdef HAVE_CDDB
   int            no_cddb;     /* If set the below are meaningless. */
   char          *cddb_email;  /* email to report to CDDB server. */
@@ -155,21 +90,13 @@ struct arguments
 #endif
   int            no_vcd;
   uint32_t       debug_level;
-  int            silent;
   int            version_only;
+  int            silent;
   int            no_header;
   int            print_iso9660;
   source_image_t source_image;
 } opts;
      
-/* Program documentation. */
-const char doc[] = 
-  PROGRAM_NAME " -- Get information about a Compact Disk or CD image.";
-
-/* A description of the arguments we accept. */
-const char args_doc[] = "[DEVICE or DISK-IMAGE]";
-
-
 /* Configuration option codes */
 enum {
   
@@ -187,111 +114,96 @@ enum {
 
 char *temp_str;
 
-struct poptOption optionsTable[] = {
-  {"debug",       'd', POPT_ARG_INT, &opts.debug_level, 0,
-   "Set debugging to LEVEL"},
-  
-  {"quiet",       'q', POPT_ARG_NONE, &opts.silent, 0,
-   "Don't produce warning output" },
-  
-  {"no-tracks",   'T', POPT_ARG_NONE, &opts.no_tracks, 0,
-   "Don't show track information"},
-  
-  {"no-analyze",  'A', POPT_ARG_NONE, &opts.no_analysis, 0,
-   "Don't filesystem analysis"},
-  
+
+/* Parse a all options. */
+static bool
+parse_options (int argc, const char *argv[])
+{
+  int opt;
+
+  struct poptOption optionsTable[] = {
+    {"access-mode",       'a', POPT_ARG_STRING, &opts.access_mode, 0,
+     "Set CD access methed"},
+    
+    {"debug",       'd', POPT_ARG_INT, &opts.debug_level, 0,
+     "Set debugging to LEVEL"},
+    
+    {"no-tracks",   'T', POPT_ARG_NONE, &opts.no_tracks, 0,
+     "Don't show track information"},
+    
+    {"no-analyze",  'A', POPT_ARG_NONE, &opts.no_analysis, 0,
+     "Don't filesystem analysis"},
+    
 #ifdef HAVE_CDDB
   {"no-cddb",     'a', POPT_ARG_NONE, &opts.no_cddb, 0,
    "Don't look up audio CDDB information or print that"},
-
-  {"cddb-port",   'P', POPT_ARG_INT, &opts.cddb_port, 0,
-   "CDDB port number to use (default 8880)"},
-
-  {"cddb-http",   'H', POPT_ARG_NONE, &opts.cddb_http, 0,
-   "Lookup CDDB via HTTP proxy (default no proxy)"},
-
-  {"cddb-server", '\0', POPT_ARG_STRING, &opts.cddb_server, 0,
-   "CDDB server to contact for information (default: freedb.freedb.org)"},
-
-  {"cddb-cache",  '\0', POPT_ARG_STRING, &opts.cddb_cachedir, 0,
-   "Location of CDDB cache directory (default ~/.cddbclient)"},
-
-  {"cddb-email",  '\0', POPT_ARG_STRING, &opts.cddb_email, 0,
-   "Email address to give CDDB server (default me@home"},
-
-  {"no-cddb-cache", '\0', POPT_ARG_NONE, &opts.cddb_disable_cache, 0,
-   "Lookup CDDB via HTTP proxy (default no proxy)"},
-
-  {"cddb-timeout",  '\0', POPT_ARG_INT, &opts.cddb_timeout, 0,
-   "CDDB timeout value in seconds (default 10 seconds)"},
-
+    
+    {"cddb-port",   'P', POPT_ARG_INT, &opts.cddb_port, 8880,
+     "CDDB port number to use (default 8880)"},
+    
+    {"cddb-http",   'H', POPT_ARG_NONE, &opts.cddb_http, 0,
+     "Lookup CDDB via HTTP proxy (default no proxy)"},
+    
+    {"cddb-server", '\0', POPT_ARG_STRING, &opts.cddb_server, 0,
+     "CDDB server to contact for information (default: freedb.freedb.org)"},
+    
+    {"cddb-cache",  '\0', POPT_ARG_STRING, &opts.cddb_cachedir, 0,
+     "Location of CDDB cache directory (default ~/.cddbclient)"},
+    
+    {"cddb-email",  '\0', POPT_ARG_STRING, &opts.cddb_email, 0,
+     "Email address to give CDDB server (default me@home"},
+    
+    {"no-cddb-cache", '\0', POPT_ARG_NONE, &opts.cddb_disable_cache, 0,
+     "Lookup CDDB via HTTP proxy (default no proxy)"},
+    
+    {"cddb-timeout",  '\0', POPT_ARG_INT, &opts.cddb_timeout, 0,
+     "CDDB timeout value in seconds (default 10 seconds)"},
 #endif
   
 #ifdef HAVE_VCDINFO
-  {"no-vcd",   'v', POPT_ARG_NONE, &opts.no_vcd, 0,
-   "Don't look up Video CD information"},
+    {"no-vcd",   'v', POPT_ARG_NONE, &opts.no_vcd, 0,
+     "Don't look up Video CD information"},
 #else 
-  {"no-vcd",   'v', POPT_ARG_NONE, &opts.no_vcd, 0,
-   "Don't look up Video CD information - for this build, this is always set"},
+    {"no-vcd",   'v', POPT_ARG_NONE, &opts.no_vcd, 1,
+     "Don't look up Video CD information - for this build, this is always set"},
 #endif
-  {"no-ioctl",  'I', POPT_ARG_NONE,  &opts.no_ioctl, 0,
-   "Don't show ioctl() information"},
-  
-  {"bin-file", 'b', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
-   OP_SOURCE_BIN, "set \"bin\" CD-ROM disk image file as source", "FILE"},
-  
-  {"cue-file", 'c', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
-   OP_SOURCE_CUE, "set \"cue\" CD-ROM disk image file as source", "FILE"},
-  
-  {"input", 'i', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
-   OP_SOURCE_AUTO,
-   "set source and determine if \"bin\" image or device", "FILE"},
-  
-  {"iso9660",  '\0', POPT_ARG_NONE, &opts.print_iso9660, 0,
-   "print directory contents of any ISO-9660 filesystems"},
+    {"no-ioctl",  'I', POPT_ARG_NONE,  &opts.no_ioctl, 0,
+     "Don't show ioctl() information"},
+    
+    {"bin-file", 'b', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
+     OP_SOURCE_BIN, "set \"bin\" CD-ROM disk image file as source", "FILE"},
+    
+    {"cue-file", 'c', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
+     OP_SOURCE_CUE, "set \"cue\" CD-ROM disk image file as source", "FILE"},
+    
+    {"input", 'i', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
+     OP_SOURCE_AUTO,
+     "set source and determine if \"bin\" image or device", "FILE"},
+    
+    {"iso9660",  '\0', POPT_ARG_NONE, &opts.print_iso9660, 0,
+     "print directory contents of any ISO-9660 filesystems"},
+    
+    {"cdrom-device", 'C', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
+     OP_SOURCE_DEVICE,
+     "set CD-ROM device as source", "DEVICE"},
+    
+    {"no-header", '\0', POPT_ARG_NONE, &opts.no_header, 
+     0, "Don't display header and copyright (for regression testing)"},
+    
+    {"nrg-file", 'N', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
+     OP_SOURCE_NRG, "set Nero CD-ROM disk image file as source", "FILE"},
+    
+    {"quiet",       'q', POPT_ARG_NONE, &opts.silent, 0,
+     "Don't produce warning output" },
+    
+    {"version", 'V', POPT_ARG_NONE, &opts.version_only, 0,
+     "display version and copyright information and exit"},
+    POPT_AUTOHELP {NULL, 0, 0, NULL, 0}
+  };
+  poptContext optCon = poptGetContext (NULL, argc, argv, optionsTable, 0);
 
-  {"cdrom-device", 'C', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
-   OP_SOURCE_DEVICE,
-   "set CD-ROM device as source", "DEVICE"},
-
-  {"no-header", '\0', POPT_ARG_NONE, &opts.no_header, 
-   0, "Don't display header and copyright (for regression testing)"},
-  
-  {"nrg-file", 'N', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &source_name, 
-   OP_SOURCE_NRG, "set Nero CD-ROM disk image file as source", "FILE"},
-  
-  {"quiet", 'q', POPT_ARG_NONE, &opts.silent, 0, 
-   "show only critical messages"},
-  
-  {"version", 'V', POPT_ARG_NONE, &opts.version_only, 0,
-   "display version and copyright information and exit"},
-  POPT_AUTOHELP {NULL, 0, 0, NULL, 0}
-};
-
-#define DEV_PREFIX "/dev/"
-static char *
-fillout_device_name(const char *device_name) 
-{
-#if defined(HAVE_WIN32_CDROM)
-  return strdup(device_name);
-#else
-  unsigned int prefix_len=strlen(DEV_PREFIX);
-  if (0 == strncmp(device_name, DEV_PREFIX, prefix_len))
-    return strdup(device_name);
-  else {
-    char *full_device_name=malloc(strlen(device_name)+prefix_len);
-    sprintf(full_device_name, DEV_PREFIX "%s", device_name);
-    return full_device_name;
-  }
-#endif
-}
-
-
-/* Parse a single option. */
-static bool
-parse_options (poptContext optCon)
-{
-  int opt;
+  program_name = strrchr(argv[0],'/');
+  program_name = program_name ? program_name+1 : strdup(argv[0]);
 
   while ((opt = poptGetNextOpt (optCon)) != -1) {
     switch (opt) {
@@ -359,46 +271,6 @@ parse_options (poptContext optCon)
   return true;
 }
 
-static void
-print_version (bool version_only)
-{
-  
-  driver_id_t driver_id;
-
-  if (!opts.no_header)
-    printf( _("CD Info %s (c) 2003 Gerd Knorr, Heiko Eifeldt & R. Bernstein\n"),
-	    VERSION);
-  printf( _("This is free software; see the source for copying conditions.\n\
-There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A\n\
-PARTICULAR PURPOSE.\n\
-"));
-
-  if (version_only) {
-    char *default_device;
-    for (driver_id=DRIVER_UNKNOWN+1; driver_id<=CDIO_MAX_DRIVER; driver_id++) {
-      if (cdio_have_driver(driver_id)) {
-	printf("Have driver: %s\n", cdio_driver_describe(driver_id));
-      }
-    }
-    default_device=cdio_get_default_device(NULL);
-    if (default_device)
-      printf("Default CD-ROM device: %s\n", default_device);
-    else
-      printf("No CD-ROM device found.\n");
-    exit(100);
-  }
-  
-}
-
-static void 
-myexit(CdIo *cdio, int rc) 
-{
-  if (NULL != cdio) 
-    cdio_destroy(cdio);
-  exit(rc);
-}
-
-
 /* ------------------------------------------------------------------------ */
 /* CDDB                                                                     */
 
@@ -610,53 +482,66 @@ print_iso9660_recurse (CdIo *cdio, const char pathname[], cdio_fs_anal_t fs,
     
   printf ("%s:\n", pathname);
 
-  cdio_assert (entlist != NULL);
+  if (NULL == entlist) {
+    fprintf (stderr, "Error getting above directory information\n");
+    return;
+  }
 
-  /* just iterate */
+  /* Iterate over files in this directory */
   
   _CDIO_LIST_FOREACH (entnode, entlist)
     {
-      char *_name = _cdio_list_node_data (entnode);
+      char *iso_name = _cdio_list_node_data (entnode);
       char _fullname[4096] = { 0, };
       iso9660_stat_t statbuf;
+      char translated_name[MAX_ISONAME+1];
+#define DATESTR_SIZE 30
+      char date_str[DATESTR_SIZE];
 
-      snprintf (_fullname, sizeof (_fullname), "%s%s", pathname, _name);
+      iso9660_name_translate(iso_name, translated_name);
+      
+      snprintf (_fullname, sizeof (_fullname), "%s%s", pathname, 
+		iso_name);
   
-      if (iso9660_fs_stat (cdio, _fullname, &statbuf, is_mode2))
-        cdio_assert_not_reached ();
+      if (iso9660_fs_stat (cdio, _fullname, &statbuf, is_mode2)) {
+        fprintf(stderr, "Error getting file info for directory %s/\n", 
+		_fullname);
+	continue;
+      }
 
       strncat (_fullname, "/", sizeof (_fullname));
 
       if (statbuf.type == _STAT_DIR
-          && strcmp (_name, ".") 
-          && strcmp (_name, ".."))
+          && strcmp (iso_name, ".") 
+          && strcmp (iso_name, ".."))
         _cdio_list_append (dirlist, strdup (_fullname));
 
       if (fs & CDIO_FS_ANAL_XA) {
-	printf ( "  %c %s %d %d [fn %.2d] [LSN %6d] ",
+	printf ( "  %c %s %d %d [fn %.2d] [LSN %6lu] ",
 		 (statbuf.type == _STAT_DIR) ? 'd' : '-',
 		 iso9660_get_xa_attr_str (statbuf.xa.attributes),
 		 uint16_from_be (statbuf.xa.user_id),
 		 uint16_from_be (statbuf.xa.group_id),
 		 statbuf.xa.filenum,
-		 statbuf.lsn);
+		 (long unsigned int) statbuf.lsn);
 	
 	if (uint16_from_be(statbuf.xa.attributes) & XA_ATTR_MODE2FORM2) {
-	  printf ("%9d (%9d)",
-		  statbuf.secsize * M2F2_SECTOR_SIZE,
-		  statbuf.size);
+	  printf ("%9u (%9u)",
+		  (unsigned int) statbuf.secsize * M2F2_SECTOR_SIZE,
+		  (unsigned int) statbuf.size);
 	} else {
-	  printf ("%9d", statbuf.size);
+	  printf ("%9u", (unsigned int) statbuf.size);
 	}
       }
-      printf ("  %s\n", _name);
+      strftime(date_str, DATESTR_SIZE, "%b %e %Y %H:%M ", &statbuf.tm);
+      printf (" %s %s\n", date_str, translated_name);
     }
 
   _cdio_list_free (entlist, true);
 
   printf ("\n");
 
-  /* now recurse */
+  /* Now recurse over the directories. */
 
   _CDIO_LIST_FOREACH (entnode, dirlist)
     {
@@ -699,7 +584,7 @@ print_iso9660_fs (CdIo *cdio, cdio_fs_anal_t fs, track_format_t track_format)
     const lsn_t extent = iso9660_get_root_lsn(&pvd);
     
     printf ("ISO9660 filesystem\n");
-    printf (" root dir in PVD set to lsn %d\n\n", extent);
+    printf (" root dir in PVD set to lsn %lu\n\n", (long unsigned) extent);
     
     print_iso9660_recurse (cdio, "/", fs, is_mode2);
   }
@@ -714,7 +599,7 @@ print_analysis(int ms_offset, cdio_analysis_t cdio_analysis,
   int need_lf;
   
   switch(CDIO_FSTYPE(fs)) {
-  case CDIO_FS_NO_DATA:
+  case CDIO_FS_AUDIO:
     if (num_audio > 0) {
       printf("Audio CD, CDDB disc ID is %08lx\n", 
 	     cddb_discid(cdio, num_tracks));
@@ -788,45 +673,28 @@ print_analysis(int ms_offset, cdio_analysis_t cdio_analysis,
     need_lf += printf("CD-Plus/Extra   ");
   if (fs & CDIO_FS_ANAL_BOOTABLE)
     need_lf += printf("bootable CD   ");
-  if (fs & CDIO_FS_ANAL_VIDEOCDI && num_audio == 0) {
+  if (fs & CDIO_FS_ANAL_VIDEOCD && num_audio == 0) {
     need_lf += printf("Video CD   ");
-#ifdef HAVE_VCDINFO
-    if (!opts.no_vcd) {
-      printf("\n");
-      print_vcd_info();
-    }
-#endif    
   }
   if (fs & CDIO_FS_ANAL_SVCD)
     need_lf += printf("Super Video CD (SVCD) or Chaoji Video CD (CVD)");
   if (fs & CDIO_FS_ANAL_CVD)
     need_lf += printf("Chaoji Video CD (CVD)");
   if (need_lf) printf("\n");
+#ifdef HAVE_VCDINFO
+  if (fs & (CDIO_FS_ANAL_VIDEOCD|CDIO_FS_ANAL_CVD|CDIO_FS_ANAL_SVCD)) 
+    if (!opts.no_vcd) {
+      printf("\n");
+      print_vcd_info();
+    }
+#endif    
+
 }
 
-/* ------------------------------------------------------------------------ */
-
-int
-main(int argc, const char *argv[])
+/* Initialize global variables. */
+static void 
+init(void) 
 {
-
-  CdIo          *cdio=NULL;
-  cdio_fs_anal_t fs=0;
-  int i;
-  lsn_t          start_track;          /* first sector of track */
-  lsn_t          data_start =0;        /* start of data area */
-  int            ms_offset = 0;
-  track_t        num_tracks=0;
-  track_t        first_track_num  =  0;
-  unsigned int   num_audio   =  0;  /* # of audio tracks */
-  unsigned int   num_data    =  0;  /* # of data tracks */
-  int            first_data  = -1;  /* # of first data track */
-  int            first_audio = -1;  /* # of first audio track */
-  cdio_analysis_t cdio_analysis; 
-  poptContext optCon = poptGetContext (NULL, argc, argv, optionsTable, 0);
-      
-  memset(&cdio_analysis, 0, sizeof(cdio_analysis));
-
   gl_default_cdio_log_handler = cdio_log_set_handler (_log_handler);
 #ifdef HAVE_CDDB
   gl_default_cddb_log_handler = cddb_log_set_handler (_log_handler);
@@ -836,9 +704,6 @@ main(int argc, const char *argv[])
   gl_default_vcd_log_handler = vcd_log_set_handler (_log_handler);
 #endif
 
-  program_name = strrchr(argv[0],'/');
-  program_name = program_name ? program_name+1 : strdup(argv[0]);
-
   /* Default option values. */
   opts.silent        = false;
   opts.no_header     = false;
@@ -858,13 +723,37 @@ main(int argc, const char *argv[])
   opts.no_ioctl      = 0;
   opts.no_analysis   = 0;
   opts.source_image  = IMAGE_UNKNOWN;
-     
+}
+
+/* ------------------------------------------------------------------------ */
+
+int
+main(int argc, const char *argv[])
+{
+
+  CdIo          *cdio=NULL;
+  cdio_fs_anal_t fs=CDIO_FS_AUDIO;
+  int i;
+  lsn_t          start_track_lsn;      /* lsn of first track */
+  lsn_t          data_start =0;        /* start of data area */
+  int            ms_offset = 0;
+  track_t        num_tracks=0;
+  track_t        first_track_num  =  0;
+  unsigned int   num_audio   =  0;  /* # of audio tracks */
+  unsigned int   num_data    =  0;  /* # of data tracks */
+  int            first_data  = -1;  /* # of first data track */
+  int            first_audio = -1;  /* # of first audio track */
+  cdio_analysis_t cdio_analysis; 
+  char           *media_catalog_number;  
+      
+  memset(&cdio_analysis, 0, sizeof(cdio_analysis));
+  init();
 
   /* Parse our arguments; every option seen by `parse_opt' will
      be reflected in `arguments'. */
-  parse_options(optCon);
+  parse_options(argc, argv);
      
-  print_version(opts.version_only);
+  print_version(program_name, VERSION, opts.no_header, opts.version_only);
 
   switch (opts.source_image) {
   case IMAGE_UNKNOWN:
@@ -909,10 +798,6 @@ main(int argc, const char *argv[])
     err_exit("%s: Error in opening device driver\n", program_name);
   } 
 
-  if (opts.debug_level > 0) {
-    printf("CD driver name: %s\n", cdio_get_driver_name(cdio));
-  }
-  
   if (source_name==NULL) {
     source_name=strdup(cdio_get_arg(cdio, "source"));
     if (NULL == source_name) {
@@ -920,6 +805,15 @@ main(int argc, const char *argv[])
     }
   } 
 
+  if (opts.silent == 0) {
+    printf("CD location   : %s\n", source_name);
+    printf("CD driver name: %s\n", cdio_get_driver_name(cdio));
+  }
+  
+  if (opts.access_mode!=NULL) {
+    cdio_set_arg(cdio, "access-mode", opts.access_mode);
+  } 
+
   first_track_num = cdio_get_first_track_num(cdio);
   num_tracks      = cdio_get_num_tracks(cdio);
 
@@ -927,7 +821,7 @@ main(int argc, const char *argv[])
     printf(STRONG "CD-ROM Track List (%i - %i)\n" NORMAL, 
 	   first_track_num, num_tracks);
 
-    printf("  #: MSF       LSN     Type\n");
+    printf("  #: MSF       LSN     Type  Green?\n");
   }
   
   /* Read and possibly print track information. */
@@ -940,18 +834,18 @@ main(int argc, const char *argv[])
 
     if (i == CDIO_CDROM_LEADOUT_TRACK) {
       if (!opts.no_tracks)
-	printf("%3d: %2.2x:%2.2x:%2.2x  %06d  leadout\n",
+	printf("%3d: %2.2x:%2.2x:%2.2x  %06lu  leadout\n",
 	       (int) i, 
 	       msf.m, msf.s, msf.f, 
-	       cdio_msf_to_lsn(&msf));
+	       (long unsigned int) cdio_msf_to_lsn(&msf));
       break;
     } else if (!opts.no_tracks) {
-      printf("%3d: %2.2x:%2.2x:%2.2x  %06d  %s\n",
+      printf("%3d: %2.2x:%2.2x:%2.2x  %06lu  %-5s %s\n",
 	     (int) i, 
 	     msf.m, msf.s, msf.f, 
-	     cdio_msf_to_lsn(&msf),
-	     track_format2str[cdio_get_track_format(cdio, i)]);
-
+	     (long unsigned int) cdio_msf_to_lsn(&msf),
+	     track_format2str[cdio_get_track_format(cdio, i)],
+	     cdio_get_track_green(cdio, i)? "true" : "false");
     }
     
     if (TRACK_FORMAT_AUDIO == cdio_get_track_format(cdio, i)) {
@@ -965,19 +859,21 @@ main(int argc, const char *argv[])
     if (i == num_tracks) i = CDIO_CDROM_LEADOUT_TRACK-1;
   }
 
+  /* get MCN */
+  media_catalog_number = cdio_get_mcn(cdio);
+  printf("Media Catalog Number (MCN): "); fflush(stdout);
+  if (NULL == media_catalog_number)
+    printf("not available\n");
+  else {
+    printf("%s\n", media_catalog_number);
+    free(media_catalog_number);
+  }
+  
+    
 #if CDIO_IOCTL_FINISHED
   if (!opts.no_ioctl) {
     printf(STRONG "What ioctl's report...\n" NORMAL);
   
-#ifdef CDROM_GET_MCN
-    /* get mcn */
-    printf("Get MCN     : "); fflush(stdout);
-    if (ioctl(filehandle,CDROM_GET_MCN, &mcn))
-      printf("FAILED\n");
-    else
-      printf("%s\n",mcn.medium_catalog_number);
-#endif
-    
 #ifdef CDROM_DISC_STATUS
     /* get disk status */
     printf("disc status : "); fflush(stdout);
@@ -1042,18 +938,18 @@ main(int argc, const char *argv[])
       
       msf_t msf;
       cdio_get_track_msf(cdio, 1, &msf);
-      start_track = cdio_msf_to_lsn(&msf);
+      start_track_lsn = cdio_msf_to_lsn(&msf);
       
-      /* CD-I/Ready says start_track <= 30*75 then CDDA */
-      if (start_track > 100 /* 100 is just a guess */) {
+      /* CD-I/Ready says start_track_lsn <= 30*75 then CDDA */
+      if (start_track_lsn > 100 /* 100 is just a guess */) {
 	fs = cdio_guess_cd_type(cdio, 0, 1, &cdio_analysis);
 	if ((CDIO_FSTYPE(fs)) != CDIO_FS_UNKNOWN)
 	  fs |= CDIO_FS_ANAL_HIDDEN_TRACK;
 	else {
 	  fs &= ~CDIO_FS_MASK; /* del filesystem info */
-	  printf("Oops: %i unused sectors at start, "
+	  printf("Oops: %lu unused sectors at start, "
 		 "but hidden track check failed.\n",
-		 start_track);
+		 (long unsigned int) start_track_lsn);
 	}
       }
       print_analysis(ms_offset, cdio_analysis, fs, first_data, num_audio,
@@ -1080,24 +976,25 @@ main(int argc, const char *argv[])
 	  ;
 	}
 	
-	start_track = (i == 1) ? 0 : cdio_msf_to_lsn(&msf);
+	start_track_lsn = (i == 1) ? 0 : cdio_msf_to_lsn(&msf);
 	
 	/* save the start of the data area */
 	if (i == first_data) 
-	  data_start = start_track;
+	  data_start = start_track_lsn;
 	
 	/* skip tracks which belong to the current walked session */
-	if (start_track < data_start + cdio_analysis.isofs_size)
+	if (start_track_lsn < data_start + cdio_analysis.isofs_size)
 	  continue;
 	
-	fs = cdio_guess_cd_type(cdio, start_track, i, &cdio_analysis);
+	fs = cdio_guess_cd_type(cdio, start_track_lsn, i, &cdio_analysis);
 
 	if (i > 1) {
 	  /* track is beyond last session -> new session found */
-	  ms_offset = start_track;
-	  printf("session #%d starts at track %2i, LSN: %6i,"
+	  ms_offset = start_track_lsn;
+	  printf("session #%d starts at track %2i, LSN: %lu,"
 		 " ISO 9660 blocks: %6i\n",
-		 j++, i, start_track, cdio_analysis.isofs_size);
+		 j++, i, (unsigned long int) start_track_lsn, 
+		 cdio_analysis.isofs_size);
 	  printf("ISO 9660: %i blocks, label `%.32s'\n",
 		 cdio_analysis.isofs_size, cdio_analysis.iso_label);
 	  fs |= CDIO_FS_ANAL_MULTISESSION;
